"""This module provides support for z transforms.

Copyright 2020 Michael Hayes, UCECE

"""

from .ratfun import Ratfun
from .sym import sympify, simplify, symsymbol, AppliedUndef
from .utils import factor_const, scale_shift
from .functions import UnitImpulse, unitimpulse, UnitStep
import sympy as sym

__all__ = ('ZT', 'IZT')

ztransform_cache = {}
inverse_ztransform_cache = {}


def ztransform_func(expr, n, z, inverse=False):

    if not isinstance(expr, AppliedUndef):
        raise ValueError('Expecting function for %s' % expr)

    scale, shift = scale_shift(expr.args[0], n)    

    zsym = sympify(str(z))
    
    # Convert v[n] to V(z), etc.
    name = expr.func.__name__
    if inverse:
        func = name[0].lower() + name[1:] + '(%s)' % z
    else:
        func = name[0].upper() + name[1:] + '(%s)' % z

    if not scale.is_constant():
        raise ValueError('Cannot determine if time-expansion or decimation')

    if scale == 1:
        result = sympify(func).subs(zsym, z)

        if shift != 0:
            result = result * z ** shift
        return result        
    
    if scale.is_integer:
        raise ValueError('Cannot do decimation yet')

    if not scale.is_rational:
        raise ValueError('Cannot handle arbitrary scaling')

    if scale.p != 1:
        raise ValueError('Cannot handle non-integer time-expansion')        
    
    result = sympify(func).subs(zsym, z ** scale.q)

    if shift != 0:
        result = result * z ** shift
    return result


def ztransform_sum(expr, n, z):

    const, expr = factor_const(expr, n)

    if len(expr.args) != 2:
        raise ValueError('Cannot compute z-transform of %s' % expr)

    if not isinstance(expr, sym.Sum):
        raise ValueError('Cannot compute z-transform of %s' % expr)

    # Look for integration of function
    if (isinstance(expr.args[0], AppliedUndef)
        and expr.args[1][0] == expr.args[0].args[0]
        and expr.args[1][1] == -sym.oo
        and expr.args[1][2] == n):
        return ztransform_func(expr.args[0].subs(expr.args[0].args[0], n), n, z) / (1 - 1 / z)
    
    # Look for convolution sum
    
    var = expr.args[1][0]
    if (expr.args[1][1] != -sym.oo) or (expr.args[1][2] != sym.oo):
        raise ValueError('Need indefinite limits for %s' % expr)
    
    const2, expr = factor_const(expr.args[0], n)
    if ((len(expr.args) != 2)
        or (not isinstance(expr.args[0], AppliedUndef))
        or (not isinstance(expr.args[1], AppliedUndef))):
        raise ValueError('Need sum of two functions: %s' % expr)        

    f1 = expr.args[0]
    f2 = expr.args[1]    

    # TODO: apply similarity theorem if have f(a tau) etc.
    
    if ((f1.args[0] != var or f2.args[0] != n - var)
        and (f2.args[0] != var or f1.args[0] != n - var)):
        raise ValueError('Cannot recognise convolution: %s' % expr)

    zsym = sympify(str(z))
    
    name = f1.func.__name__
    func1 = name[0].upper() + name[1:] + '(%s)' % str(zsym)

    name = f2.func.__name__
    func2 = name[0].upper() + name[1:] + '(%s)' % str(zsym)    

    F1 = sympify(func1).subs(zsym, z)
    F2 = sympify(func2).subs(zsym, z)
    
    return F1 * F2


def remove_heaviside(expr, n):

    rest = sym.S.One
    for factor in expr.as_ordered_factors():
        if (factor.is_Function and factor.func in (sym.Heaviside, UnitStep) and
            factor.args[0] == n):
            # Could remove Heaviside[n+m] where m > 0
            pass
        else:
            rest *= factor
    return rest


def ztransform_term(expr, n, z):

    const, expr = factor_const(expr, n)

    expr = remove_heaviside(expr, n)
    
    if expr.has(sym.Sum):
        try:
            return ztransform_sum(expr, n, z) * const
        except:
            pass

    nsym = sympify(str(n))
    expr = expr.replace(nsym, n)

    # foo = 1 / (1 - 1 / z)
    # factors = expr.as_ordered_factors()
    # if foo in factors:
    #     could remove factor, find ZT of rest, and then integrate....
    
    if expr.has(AppliedUndef):

        rest = sym.S.One
        expr = expr.cancel()
        for factor in expr.as_ordered_factors():
            if isinstance(factor, AppliedUndef):
                result = ztransform_func(factor, n, z)
            else:
                if factor.has(n):
                    raise ValueError('TODO: need derivative of undefined'
                                     ' function for %s' % factor)
                rest *= factor
        return result * rest * const

    invz = z ** -1

    result = None
    args = expr.args
    
    if expr.is_Function and expr.func == UnitImpulse:
        if args[0] is n:
            result = 1
        delay = n - args[0]
        if not delay.has(n):
            result = invz ** delay

    elif expr == 1:
        # Unilateral ZT
        result = 1 / (1 - invz)
        
    elif expr.is_Function and expr.func in (sym.Heaviside, UnitStep):
        if args[0] is n:
            result = 1 / (1 - invz)
        else:
            delay = n - args[0]
            if not delay.has(n):
                result = invz ** delay * 1 / (1 - invz)

    elif expr.is_Function and expr.func == sym.cos:
        aconst, aexpr = factor_const(args[0], n)
        if aexpr == n:
            result = (1 - sym.cos(aconst) * invz) / (1 - 2 * sym.cos(aconst) * invz + invz ** 2)

    elif expr.is_Function and expr.func == sym.sin:
        aconst, aexpr = factor_const(args[0], n)
        if aexpr == n:
            result = (sym.sin(aconst) * invz) / (1 - 2 * sym.cos(aconst) * invz + invz ** 2)            

    # Handle a**n
    elif expr.is_Pow and expr.args[1] == n:
        # TODO, handle n + m, etc.
        result = 1 / (1 - expr.args[0] * invz)

    # Handle a**(-n)
    elif (expr.is_Pow and expr.args[1].is_Mul and
          expr.args[1].args[0] == -1 and expr.args[1].args[1] == n):
        result = 1 / (1 - (1 / expr.args[0]) * invz)          
            
    if result is None:
        # Use m instead of n to avoid n and z in same expr.
        # TODO, check if m already used...
        msym = sympify('m', real=True)        
        nsym = sympify(str(n))        
        zsym = sympify(str(z))
        result = sym.Sum(expr.subs(nsym, msym) * zsym**msym, (msym, 0, sym.oo))
        
    return const * result


def ztransform(expr, n, z, evaluate=True):
    """Compute unilateral z-transform of expr with lower index 0.

    Undefined functions such as v[n] are converted to V(z)

    """

    if expr.is_Equality:
        return sym.Eq(ztransform(expr.args[0], n, z),
                      ztransform(expr.args[1], n, z))

    if not evaluate:
        result = sym.Sum(expr * z**(-n), (n, 0, sym.oo))
        return result
    
    const, expr = factor_const(expr, n)    
    key = (expr, n, z)
    if key in ztransform_cache:
        return const * ztransform_cache[key]

    if expr.has(z):
        raise ValueError('Cannot Z transform expression %s that depends on %s' % (expr, z))
    
    # The variable may have been created with different attributes,
    # say when using sympify('Heaviside(n)') since this will
    # default to assuming that n is complex.  So if the symbol has the
    # same representation, convert to the desired one.

    var = sym.Symbol(str(n))
    if isinstance(expr, Expr):
        expr = expr.expr
    else:
        expr = sympify(expr)

    # SymPy ztransform barfs on Piecewise but unilateral ZT ignores expr
    # for n < 0 so remove Piecewise.
    expr = expr.replace(var, n)        
    if expr.is_Piecewise and expr.args[0].args[1].has(n >= 0):
        expr = expr.args[0].args[0]

    expr = sym.expand(expr)        
    terms = expr.as_ordered_terms()
    result = 0

    try:
        for term in terms:
            result += ztransform_term(term, n, z)
    except ValueError:
        raise

    result = result.simplify()
    ztransform_cache[key] = result
    return (const * result).simplify()


def inverse_ztransform_ratfun(expr, z, n, **assumptions):

    expr = expr / z
    
    # Handle special case 1 / (z**m * (z - 1)) since this becomes u[n - m]
    # The default method produces u[n] - delta[n] for u[n-1].  This is correct
    # but can be simplified.
    # In general, 1 / (z**m * (z - a)) becomes a**n * u[n - m]

    if (len(expr.args) == 2 and expr.args[1].is_Pow and
        expr.args[1].args[0].is_Add and
        expr.args[1].args[0].args[0] == -1 and
        expr.args[1].args[0].args[1] == z):

        delay = None
        if expr.args[0] == z:
            delay = 1
        elif expr.args[0].is_Pow and expr.args[0].args[0] == z:
            a = expr.args[0].args[1]
            if a.is_positive:
                print('Warning, dodgy z-transform.  Have advance of unit step.')
            elif not a.is_negative:
                print('Warning, dodgy z-transform.  May have advance of unit step.')                
            delay = -a
        elif (expr.args[0].is_Pow and expr.args[0].args[0].is_Pow and
              expr.args[0].args[0].args[0] == z and
              expr.args[0].args[0].args[1] == -1):              
            a = expr.args[0].args[1]
            if a.is_negative:
                print('Warning, dodgy z-transform.  Have advance of unit step.')
            elif not a.is_positive:
                print('Warning, dodgy z-transform.  May have advance of unit step.')                
            delay = a            

        if delay is not None:
            return UnitStep(n - delay), sym.S.Zero
        
    damping = assumptions.get('damping', None)
    
    zexpr = Ratfun(expr, z)

    Q, M, D, delay, undef = zexpr.as_QMD()

    cresult = sym.S.Zero
    uresult = sym.S.Zero

    if Q:
        Qpoly = sym.Poly(Q, z)        
        C = Qpoly.all_coeffs()
        for m, c in enumerate(C):
            cresult += c * unitimpulse(n - len(C) + m + 1)

    # There is problem with determining residues if
    # have 1/(z*(-a/z + 1)) instead of 1/(-a + z).  Hopefully,
    # simplify will fix things...
    expr = (M / D).simplify()
    for factor in expr.as_ordered_factors():
        if factor == sym.oo:
            return factor, factor

    zexpr = Ratfun(expr, z)
    poles = zexpr.poles(damping=damping)
    polesdict = {}
    for pole in poles:
        polesdict[pole.expr] = pole.n
    
    for pole in poles:

        p = pole.expr

        # Number of occurrences of the pole.
        o = polesdict[p]        

        if o == 0:
            continue

        if o == 1:
            r = zexpr.residue(p, poles)

            # TODO combine conjugates to get a real result
            # See laplace.py

            if p == 0:
                cresult += r * unitimpulse(n)
            else:
                uresult += r * p ** n
            continue

        # Handle repeated poles.
        
        expr2 = expr * (z - p) ** o
        expr2 = expr2.simplify()
        for i in range(1, o + 1):
            m = o - i
            r = sym.limit(
                sym.diff(expr2, z, m), z, p) / sym.factorial(m)

            if p == 0:
                cresult += r * unitimpulse(n - i + 1)
            else:            
                uresult += r * (p ** n) * n**(i - 1)

    # cresult is a sum of Dirac deltas and its derivatives so is known
    # to be causal.

    return cresult, uresult


def dummyvar(intnum=0):
    if intnum == 0:
        return sympify('m', real=True)
    else:
        return sympify('m_%d' % intnum, real=True)


def inverse_ztransform_product(expr, z, n, **assumptions):

    # Handle expressions with a function of z, e.g., V(z) * Y(z), V(z)
    # / z etc.

    zsym = sympify(str(z))    

    if assumptions.get('causal', False):
        # Assume that all functions are causal in the expression.
        n1 = sym.S.Zero
        n2 = n
    else:
        n1 = -sym.oo
        n2 = sym.oo        
    
    const, expr = factor_const(expr, z)

    factors = expr.as_ordered_factors()
    if len(factors) < 2:
        raise ValueError('Expression does not have multiple factors: %s' % expr)

    if (len(factors) == 3 and factors[0] == z and
        isinstance(factors[2], AppliedUndef) and
        factors[1].is_Pow and factors[1].args[1] == -1 and
        factors[1].args[0].is_Add and factors[1].args[0].args[0] == -1
        and factors[1].args[0].args[1] == z):
        # Handle cumulative sum  z / (z - 1) * V(z)
        m = dummyvar()
        result = ztransform_func(factors[2], z, m, True)                
        result = sym.Sum(result, (m, n1, n))
        return result
    
    # TODO, is this useful?
    if (len(factors) > 2 and not
        isinstance(factors[1], AppliedUndef) and
        isinstance(factors[2], AppliedUndef)):
        factors = [factors[0], factors[2], factors[1]] + factors[3:]

    # TODO, is this useful?        
    if isinstance(factors[1], AppliedUndef):
        terms = factors[0].as_ordered_terms()
        if len(terms) >= 2:
            result = sym.S.Zero
            for term in terms:
                result += inverse_ztransform_product(factors[1] * term, z, n)
            return const * result

    cresult, uresult = inverse_ztransform_term1(factors[0], z, n)
    result = cresult + uresult

    intnum = 0
    for m in range(len(factors) - 1):
        if m == 0 and isinstance(factors[1], AppliedUndef):
            # Note, as_ordered_factors puts powers of z before the functions.
            if factors[0] == z:
                # Handle time-advance
                # Convert z * V(z) to v[n + 1]
                # TODO, fix for unilateral ZT 
                result = ztransform_func(factors[1], z, n, True)            
                result = result.subs(n, n + 1)
                continue
            elif factors[0].is_Pow and factors[0].args[0] == z and factors[0].args[1] > 0:
                # Handle higher order advances
                # Convert z ** k * V(z) to v[n + k]
                result = ztransform_func(factors[1], z, n, True)            
                result = result.subs(n, n + factors[0].args[1])
                continue                
            elif factors[0].is_Pow and factors[0].args[0] == z and factors[0].args[1] < 0:
                # Handle time-delay  1 / z ** k * V(z)
                result = ztransform_func(factors[1], z, n, True)
                result = result.subs(n, n + factors[0].args[1])                
                continue
            elif (factors[0].is_Pow and
                  factors[0].args[0].is_Add and
                  factors[0].args[1] == -1 and                  
                  factors[0].args[0].args[0] == 1 and
                  factors[0].args[0].args[1].is_Mul and
                  factors[0].args[0].args[1].args[0] == -1 and
                  factors[0].args[0].args[1].args[1].is_Pow and
                  factors[0].args[0].args[1].args[1].args[1] == -1 and
                  factors[0].args[0].args[1].args[1].args[0] is zsym):
                # Handle cumulative sum  1 / (1 - 1 / z) * V(z)
                m = dummyvar(intnum)
                result = ztransform_func(factors[1], z, m, True)                
                intnum += 1
                result = sym.Sum(result, (m, n1, n))
                continue

        # Convert product to convolution
        dummy = dummyvar(intnum)
        intnum += 1
        cresult, uresult = inverse_ztransform_term1(factors[m + 1], z, n)
        expr2 = cresult + uresult
        result = sym.Sum(result.subs(n, n - dummy) * expr2.subs(n, dummy),
                         (dummy, n1, n2))
    
    return const * result


def inverse_ztransform_power(expr, z, n, **assumptions):

    # Handle expressions with a power of z.
    if (expr.is_Pow and expr.args[0] == z):
        exponent = expr.args[1]

        if exponent.is_positive:
            print('Warning, dodgy z-transform.  Have advance of unit impulse.')
        elif not exponent.is_negative:
            print('Warning, dodgy z-transform.  May have advance of unit impulse.')
        
        return unitimpulse(n + exponent), sym.S.Zero

    # Handle expressions with a power of (1 / z).
    if (expr.is_Pow and expr.args[0].is_Pow and
        expr.args[0].args[0] == z and expr.args[0].args[1] == -1):

        exponent = expr.args[1]

        if exponent.is_negative:
            print('Warning, dodgy z-transform.  Have advance of unit impulse.')
        elif not exponent.is_positive:
            print('Warning, dodgy z-transform.  May have advance of unit impulse.')
        
        return unitimpulse(n - exponent), sym.S.Zero        

    raise ValueError('Expression %s is not a power of z' % expr)


def inverse_ztransform_sympy(expr, z, n):

    # This barfs when needing to generate Dirac deltas
    from sympy.integrals.transforms import inverse_ztransform_transform
    result = inverse_ztransform_transform(expr, n, z)
    
    if result.has(sym.InverseZtransformTransform):
        raise ValueError('Cannot determine inverse z-transform'
                         ' transform of %s with sympy' % expr)
    return result


def inverse_ztransform_term1(expr, z, n, **assumptions):

    const, expr = factor_const(expr, z)

    if expr == 1:
        return const * unitimpulse(n), sym.S.Zero
    
    if isinstance(expr, AppliedUndef):
        # Handle V(z), 3 * V(z) etc.  If causal is True it is assumed
        # that the unknown functions are causal.  Note ztransform_func
        # just changes the name so it works as inverse_ztransform_func.
        result = ztransform_func(expr, z, n, True)
        return const * result, sym.S.Zero
    
    if expr.has(AppliedUndef):
        return const * inverse_ztransform_product(expr, z, n,
                                                  **assumptions), sym.S.Zero

    if expr == z:
        print('Warning, dodgy z-transform.  Have advance of unit impulse.') 
        return const * unitimpulse(n + 1), sym.S.Zero        

    if (expr.is_Pow and
        (expr.args[0] == z or
         (expr.args[0].is_Pow and
          expr.args[0].args[0] == z and expr.args[0].args[1] == -1))):
        cresult, uresult = inverse_ztransform_power(expr, z, n, **assumptions)
        return const * cresult, const * uresult        
        
    try:
        # This is the common case.
        cresult, uresult = inverse_ztransform_ratfun(expr, z, n, **assumptions)
        return const * cresult, const * uresult
    except:
        pass

    try:
        return sym.S.Zero, const * inverse_ztransform_sympy(expr, z, n, **assumptions)
    except:
        pass

    # As last resort see if can convert to convolutions...
    return sym.S.Zero, const * inverse_ztransform_product(expr, z, n)
    
    
def inverse_ztransform_term(expr, z, n, **assumptions):

    cresult, uresult = inverse_ztransform_term1(expr, z, n, **assumptions)

    if assumptions.get('causal', False):
        uresult = uresult * UnitStep(n)
    
    return cresult, uresult


def inverse_ztransform_by_terms(expr, z, n, **assumptions):

    terms = expr.as_ordered_terms()

    cresult = sym.S.Zero
    uresult = sym.S.Zero    

    for term in terms:
        part1, part2 = inverse_ztransform_term(term, z, n, **assumptions)
        cresult += part1
        uresult += part2        
    return cresult, uresult


def inverse_ztransform_make(n, const, cresult, uresult, **assumptions):

    result = const * (cresult + uresult)
    
    if assumptions.get('dc', False):
        free_symbols = set([symbol.name for symbol in result.free_symbols])
        if 'n' in free_symbols:
            raise ValueError('Something wonky going on, expecting dc.')
    
    elif assumptions.get('ac', False):

        if cresult != 0:
            raise ValueError('Inverse z-transform weirdness for %s with is_ac True' % result)
        # TODO, perform more checking of the result.
        
    elif not assumptions.get('causal', False):

        # Cannot determine result for n < 0
        result = sym.Piecewise((result, n >= 0))
            
    return result    


def inverse_ztransform1(expr, z, n, **assumptions):

    const, expr = factor_const(expr, z)
    
    key = (expr, z, n, 
           assumptions.get('causal', False),
           assumptions.get('damping', None))
    
    if key in inverse_ztransform_cache:
        cresult, uresult = inverse_ztransform_cache[key]        
        return const, cresult, uresult        

    try:
        if expr.is_Add:
            cresult, uresult = inverse_ztransform_by_terms(expr, z, n, **assumptions)
        else:
            try:
                cresult, uresult = inverse_ztransform_term(expr, z, n, **assumptions)
            except:
                expr = sym.expand(expr)
                cresult, uresult = inverse_ztransform_by_terms(expr, z, n,
                                                               **assumptions)
    except:
        raise ValueError('Cannot determine z-transform of %s' % expr)
        
    # cresult is known to be causal, uresult is unsure
    inverse_ztransform_cache[key] = cresult, uresult
    return const, cresult, uresult    


def inverse_ztransform(expr, z, n, **assumptions):
    """Calculate inverse z-transform of X(z) and return x[n].

    The unilateral z-transform cannot determine x[n] for n < 0
    unless given additional information in the way of assumptions.

    The assumptions are:
    dc -- x[n] = constant so X(z) must have the form constant * z / (z - 1)
    causal -- x[n] = 0 for n < 0.
    ac -- x[n] = A cos(a * t) + B * sin(b * t)
    """

    if expr.is_Equality:
        return sym.Eq(inverse_ztransform(expr.args[0], z, n, **assumptions),
                      inverse_ztransform(expr.args[1], z, n, **assumptions))

    if expr.has(n):
        raise ValueError('Cannot inverse z-transform  %s that depends on %s' % (expr, n))

    const, cresult, uresult = inverse_ztransform1(expr, z, n, **assumptions)
    return inverse_ztransform_make(n, const, cresult, uresult, **assumptions)


def ZT(expr, n, z, **assumptions):
    """Compute unilateral Z-Transform transform of expr with lower limit 0.

    Undefined functions such as v[n] are converted to V(z)."""
    
    return ztransform(expr, n, z, **assumptions)


def IZT(expr, z, n, **assumptions):
    """Calculate inverse z-Transform of X(s) and return x[n].

    The unilateral Z-Transform transform cannot determine x[n] for n < 0
    unless given additional information in the way of assumptions.

    The assumptions are:
    dc -- x[n] = constant
    causal -- x[n] = 0 for n < 0.
    ac -- x[n] = A cos(a * n) + B * sin(b * n)
    """
    
    return inverse_ztransform(expr, z, n, **assumptions)

from .expr import Expr




"""This module provides the DiscreteTimeDomainExpression class to
represent discrete-time expressions.

Copyright 2020--2021 Michael Hayes, UCECE

"""

from __future__ import division
from .domains import DiscreteTimeDomain
from .seqexpr import SequenceExpression
from .sequence import Sequence
from .functions import exp
from .sym import j, oo, pi, fsym
from .dsym import nsym, ksym, zsym, dt
from .ztransform import ztransform
from .dft import DFT
from sympy import Sum, summation, limit


__all__ = ('nexpr', )

class DiscreteTimeDomainExpression(DiscreteTimeDomain, SequenceExpression):
    """Discrete time expression or symbol."""

    var = nsym


    def __init__(self, val, **assumptions):

        check = assumptions.pop('check', True)
        assumptions['real'] = True        
        
        super(DiscreteTimeDomainExpression, self).__init__(val, **assumptions)

        expr = self.expr
        if check and expr.find(zsym) != set() and not expr.has(Sum):
            raise ValueError(
                'n-domain expression %s cannot depend on z' % expr)
        if check and expr.find(ksym) != set() and not expr.has(Sum):
            raise ValueError(
                'n-domain expression %s cannot depend on k' % expr)            

    def _mul_compatible_domains(self, x):

        if self.domain == x.domain:
            return True        

        return x.is_constant_domain

    def _div_compatible_domains(self, x):

        if self.domain == x.domain:
            return True
        
        return x.is_constant_domain

    def as_expr(self):
        return DiscreteTimeDomainExpression(self)

    def differentiate(self):
        """First order difference."""

        result = (self.expr - self.subs(n - 1).expr) / dt
        return self.__class__(result, **self.assumptions)

    def integrate(self):
        """First order integration."""

        from .sym import sympify
        from .utils import factor_const
        from .functions import UnitImpulse, u

        # TODO, get SymPy to optimize this case.
        expr = self.expr
        const, expr = factor_const(expr, nsym)
        if expr.is_Function and expr.func == UnitImpulse:
            return dt * u(expr.args[0]) * const
        
        msym = sympify('m', real=True)
        result = dt * summation(self.subs(msym).expr, (msym, -oo, nsym))
        
        return self.__class__(result, **self.assumptions)

    def ztransform(self, evaluate=True, **assumptions):
        """Determine one-sided z-transform."""

        assumptions = self.assumptions.merge_and_infer(self, **assumptions)
        result = ztransform(self.expr, self.var, zsym, evaluate)
        return self.change(result, domain='Z', **assumptions)

    def ZT(self, **assumptions):
        return self.ztransform(**assumptions)

    def plot(self, n=None, **kwargs):
        """Plot the sequence.  If n is not specified, it defaults to the
        range (-20, 20).  n can be a vector of specified instants, a
        tuple specifing the range, or a constant specifying the
        maximum value with the minimum value set to 0.

        kwargs include:
        axes - the plot axes to use otherwise a new figure is created
        xlabel - the x-axis label
        ylabel - the y-axis label
        xscale - the x-axis scaling, say for plotting as ms
        yscale - the y-axis scaling, say for plotting mV
        in addition to those supported by the matplotlib plot command.
        
        The plot axes are returned."""

        if n is None:
            n = (-20, 20)
        
        from .plot import plot_sequence
        return plot_sequence(self, n, **kwargs)

    def initial_value(self):
        """Determine value at n = 0."""

        return self.subs(0)

    def final_value(self):
        """Determine value at n = oo."""

        return self.__class__(limit(self.expr, self.var, oo))

    def DFT(self, N=None, evaluate=True):

        from .kexpr import k

        if N is None:
            from .sym import sympify
            
            N = sympify('N')

        result = DFT(self.expr, nsym, ksym, N, evaluate=evaluate)
        return self.change(result, domain='discrete fourier')
    
    def delay(self,m):
        """Delay signal by m samples."""

        return self.subs(n - m)

    def extent(self, n1=-100, n2=100):
        """Determine extent of the signal.

        For example, nexpr([1, 1]).extent() = 2
                     nexpr([1, 0, 1]).extent() = 3
                     nexpr([0, 1, 0, 1]).extent() = 3

        This performs a search between n=n1 and n=n2."""

        return self.seq((n1, n2)).extent()

    def discrete_time_fourier_transform(self, **assumptions):
        """Convert to Fourier domain using discrete time Fourier transform."""

        assumptions = self.assumptions.merge_and_infer(self, **assumptions)
        if assumptions.is_causal:
            return self.ZT(**assumptions).DTFT()

        from .fexpr import fexpr

        # TODO, handle evaluation keyword
        return fexpr(summation(self.subs(nsym).expr * exp(-2 * j * pi * fsym * nsym * dt), (nsym, -oo, oo)))

    def DTFT(self, **assumptions):
        """Convert to Fourier domain using discrete time Fourier transform."""
    
        return self.discrete_time_fourier_transform(**assumptions)

    def difference_equation(self, input='x', output='y', form='iir'):
        """Create difference equation from impulse response.

        form can be 'fir' or 'iir' ('direct form I').
        """

        H = self.ZT()
        return H.difference_equation(input, output, form)


def nexpr(arg, **assumptions):
    """Create nExpr object.  If `arg` is nsym return n"""

    if arg is nsym:
        return n

    from .seq import seq
    
    if isinstance(arg, str) and arg.startswith('{'):
        return seq(arg)
    
    from numpy import ndarray
    
    if isinstance(arg, (list, ndarray)):
        return Sequence(arg, var=n).as_impulses()

    return DiscreteTimeDomainExpression(arg, **assumptions)


from .expressionclasses import expressionclasses

classs = expressionclasses.register('discrete time', DiscreteTimeDomainExpression)

n = DiscreteTimeDomainExpression('n')



import sympy as sym
from sympy.core import S, Integer
from sympy.core.logic import fuzzy_not

class UnitImpulse(sym.Function):

    is_integer = True
    
    @classmethod
    def eval(cls, nval):
        """
        Evaluates the discrete unit impulse function.
        """
        
        if nval.is_zero:
            return S.One
        elif fuzzy_not(nval.is_zero):
            return S.Zero


class UnitStep(sym.Function):

    is_integer = True
    
    @classmethod
    def eval(cls, nval):
        """
        Evaluates the discrete unit step function.
        """
        
        if nval.is_nonnegative:
            return S.One
        elif nval.is_negative:
            return S.Zero

        
class rect(sym.Function):

    @classmethod
    def eval(cls, val):
        """
        Evaluates the rectangle function.
        """

        if val.is_Number:
            if val < -0.5 or val > 0.5:
                return S.Zero
            return S.One

    def rewrite(self, *args, **hints):

        x = self.args[0]
        return sym.Heaviside(x + S.Half) - sym.Heaviside(x - S.Half)

        
class sinc(sym.Function):

    @classmethod
    def eval(cls, val):
        """
        Evaluates the sinc (cardinal sine) function.
        """

        if val.is_Number:
            if val == 0:
                return S.One
            x = sym.pi * val
            return sym.sin(x) / x
        
    def rewrite(self, *args, **hints):

        x = sym.pi * self.args[0]
        return sym.sin(x) / x
    



"""This module provides preliminary support for polyphase two-ports.
For three-phase systems, these are actually six-ports.  They extend
the concept of cascade matrices (ABCD matrices) for polyphase systems.

Copyright 2020 Michael Hayes, UCECE

"""


from .matrix import Matrix
from .expr import expr, Expr
from .polyphase import polyphase_decompose_matrix, polyphase_compose_matrix
from .polyphase import alpha_simplify3
from sympy import Add

class Polytwoport(Matrix):

    @classmethod
    def series(cls, *args, Zn=0):
        """The args are series impedances, for example:
        
         X = Polytwoport.series('Za', 'Zb', 'Zc')

        """

        args = [expr(arg) for arg in args]
        N = len(args)

        Zn = expr(Zn).expr
        
        I = Matrix.eye(N)        
        
        obj = cls(Matrix.zeros(2 * N))
        obj.A = I
        obj.B = Matrix.diag(args) + Matrix.ones(N) * Zn
        obj.D = I
        return obj

    @classmethod
    def shunt(cls, *args):
        """The args are shunt admittances to the common node, for example:
        
         X = Polytwoport.shunt('Ya', 'Yb', 'Yc')

        """

        args = [expr(arg).expr for arg in args]        
        N = len(args)

        Ysum = Add(*args)        

        I = Matrix.eye(N)        
        C = Matrix.zeros(N)

        for i in range(N):
            for j in range(N):
                if i == j:
                    C[i, j] = args[i]
                else:
                    C[i, j] = - args[i] * args[j] / Ysum
        
        obj = cls(Matrix.zeros(2 * N))
        obj.A = I
        obj.C = C
        obj.D = I
        return obj

    @classmethod
    def star(cls, *args):
        """The args are shunt admittances to the common node, for example:
        
         X = Polytwoport.star('Yas', 'Ybs', 'Ycs', 'Ysg')

        """

        args = [expr(arg).expr for arg in args]
        N = len(args) - 1

        if N != 3:
            raise ValueError('Can only handle N=3')

        I = Matrix.eye(N)

        d1 = Add(*args[0:-1])        
        d = Add(*args)
        Yas, Ybs, Ycs, Ysg = args
        
        obj = cls(Matrix.zeros(2 * N))
        obj.A = I
        obj.B = I * 0
        obj.C = Matrix((( Yas * d1,  -Yas * Ybs, -Yas * Ycs),
                        (-Yas * Ybs,  Ybs * d1,  -Ybs * Ycs),
                        (-Yas * Ycs, -Ybs * Ycs,  Ycs * d1))) / d
        obj.D = I
        return obj
        
    @property
    def N_phases(self):
        return self.shape[0] // 2

    @property
    def N(self):
        return self.N_phases

    @property
    def A(self):
        return self[0:self.N, 0:self.N]

    @property
    def B(self):
        return self[0:self.N, self.N:]

    @property
    def C(self):
        return self[self.N:, 0:self.N]

    @property
    def D(self):
        return self[self.N:, self.N:]

    @A.setter
    def A(self, A):
        self[0:self.N, 0:self.N] = A

    @B.setter
    def B(self, B):
        self[0:self.N, self.N:] = B

    @C.setter
    def C(self, C):
        self[self.N:, 0:self.N] = C

    @D.setter
    def D(self, D):
        self[self.N:, self.N:] = D

    def transform(self, W=None, Winv=None, simplify=True):

        if W is None:
            W = polyphase_decompose_matrix(self.N)
        
        if Winv is None:
            Winv = polyphase_decompose_matrix(self.N)

        new = self * 0
        
        new.A = Winv * self.A * W
        new.B = Winv * self.B * W
        new.C = Winv * self.C * W
        new.D = Winv * self.D * W

        if simplify:
            new = new.alpha_simplify()
        
        return new

    def alpha_simplify(self, alpha=None):

        if self.N != 3:
            return self
        
        return alpha_simplify3(self, alpha)



"""This module provides the ImmittanceMixin class.  It provides common
methods for immittances.

Copyright 2020 Michael Hayes, UCECE

"""

class ImmittanceMixin(object):

    @property
    def R(self):
        """Resistance."""
        from .symbols import s
        
        ret = self.Z.real
        if self.is_causal:
            ret = ret.replace(s.real, 0)
        return ret

    @property
    def resistance(self):
        """Resistance."""
        return self.R

    @property
    def X(self):
        """Reactance."""
        return self.Z.imag

    @property
    def reactance(self):
        """Reactance."""
        return self.X

    @property
    def G(self):
        """Conductance.
        
        Note Y = G + j * B = 1 / Z = 1 / (R + j * X)
        and so G = R / (R**2 + X**2).

        Thus for DC, when X = 0, then G = 1 / R and is infinite for R
        = 0.  However, if Z is purely imaginary, i.e, R = 0 then G = 0,
        not infinity as might be expected.

        """
        from .symbols import s
        
        ret = self.Y.real
        if self.is_causal:
            ret = ret.replace(s.real, 0)
        return ret        

    @property
    def conductance(self):
        """Conductance."""
        return self.G    
    
    @property
    def B(self):
        """Susceptance."""
        return -self.Y.imag

    @property
    def Y(self):
        """Admittance.

        The admittance is expressed in jomega form for AC circuits
        and in s-domain for for transient circuits.
        
        Use Y(omega) or Y(s) to achieve the desired form."""
        
        return self.admittance

    @property
    def Z(self):
        """Impedance.

        The impedance is expressed in jomega form for AC circuits
        and in s-domain for for transient circuits.  

        Use Z(omega) or Z(s) to achieve the desired form."""

        return self.impedance    

    def network(self, form='default'):
        """Synthesise a network with an equivalent impedance.
        `form` includes: cauerI, cauerII, fosterI, fosterII.

        Note some methods generate networks with negative value
        components."""

        from .synthesis import network
        
        return network(self.Z, form)



from lcapy import *
import unittest
import sympy as sym

class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def assertEqual2(self, ans1, ans2, comment):

        try:
            self.assertEqual(ans1, ans2, comment)
        except AssertionError as e:
            ans1.pprint()
            ans2.pprint()
            raise AssertionError(e)

    def test_impedance(self):

        ZC = C(3).Z

        self.assertEqual(ZC.Y(jw), 3 * jw, '.Y(jw)')
        self.assertEqual(ZC.Z(jw), -j / (3 * omega), '.Z(jw)')        
        self.assertEqual(ZC.R(jw), 0, '.R(jw)')
        self.assertEqual(ZC.G(jw), 0, '.G(jw)')
        self.assertEqual(ZC.X(jw), -1 / (3 * omega), '.X(jw)')
        self.assertEqual(ZC.B(jw), -3 * omega, '.B(jw)')

        ZR = R(3).Z

        self.assertEqual(ZR.Y(jw), expr('1 / 3'), '.Y(jw)')
        self.assertEqual(ZR.Z(jw), 3, '.Z(jw)')        
        self.assertEqual(ZR.R(jw), 3, '.R(jw)')
        self.assertEqual(ZR.G(jw), expr('1 / 3'), '.G(jw)')
        self.assertEqual(ZR.X(jw), 0, '.X(jw)')
        self.assertEqual(ZR.B(jw), 0, '.B(jw)')

        ZL = L(3).Z

        self.assertEqual(ZL.Y(jw), -j / (3 * omega), '.Y(jw)')
        self.assertEqual(ZL.Z(jw), 3 * jw, '.Z(jw)')        
        self.assertEqual(ZL.R(jw), 0, '.R(jw)')
        self.assertEqual(ZL.G(jw), 0, '.G(jw)')
        self.assertEqual(ZL.X(jw), 3 * omega, '.X(jw)')
        self.assertEqual(ZL.B(jw), 1 / (3 * omega), '.B(jw)')                

        Z1 = (R(2) + L(3)).Z

        self.assertEqual(Z1.Y(jw), 1 / (3 * jw + 2), '.Y(jw)')
        self.assertEqual(Z1.Z(jw), 3 * jw + 2, '.Z(jw)')        
        self.assertEqual(Z1.R(jw), 2, '.R(jw)')
        self.assertEqual(Z1.G(jw), 2 / (9 * omega**2 + 4), '.G(jw)')
        self.assertEqual(Z1.X(jw), 3 * omega, '.X(jw)')
        self.assertEqual(Z1.B(jw), 3 * omega / (9 * omega**2 + 4), '.B(jw)')

    def test_reciprocal(self):

        Z1 = impedance(2)
        Y1 = admittance(1 / 2)

        self.assertEqual(1 / Z1, Y1, '1 / Z')
        self.assertEqual(1 / Y1, Z1, '1 / Y')                



"""This module contains configuration information.

Copyright 2019--2020 Michael Hayes, UCECE

"""

# SymPy symbols to exclude.  It might be easier to add the ones we want...
excludes = ['I', 'C', 'O', 'S', 'N', 'E', 'E1', 'Q', 'beta', 'gamma', 'zeta',
            'Le', 'Lt', 'Ge', 'Gt', 'Ci']

# Aliases for SymPy symbols
aliases = {'delta': 'DiracDelta', 'step': 'Heaviside', 'u': 'Heaviside',
           'H': 'Heaviside', 'j': 'I'}

import sympy as sym
str_expr_map = {sym.I: 'j'}

latex_expr_map = {sym.I: '\mathrm{j}', sym.Heaviside: 'u'}

# Hack to pretty print i as j
junicode = '\u2149'
from sympy.printing.pretty.pretty_symbology import atoms_table
atoms_table['ImaginaryUnit'] = junicode

pretty_expr_map = {sym.I: junicode, sym.Heaviside: 'u'}

# Words to format in Roman font for LaTeX expressions. 

functions = ('heaviside', 'diracdelta', 'conjugate', 'sqrt', 'exp',
             'log', 'log10', 'sin', 'cos', 'tan', 'cot' 'asin',
             'acos', 'atan', 'atan2', 'acot', 'sinh', 'cosh', 'tanh', 'asinh',
             'acosh', 'atanh', 'gcd', 'abs', 'unitimpulse', 'arg', 'sign',
             'rect', 'sinc')

subscripts = ('in', 'out', 'ref', 'rms', 'load', 'source', 'avg',
              'mean', 'peak', 'pk', 'pk-pk', 'pp', 'min', 'max', 'src', 'bat',
              'cc', 'ee', 'dd', 'ss', 'ih', 'il', 'oh', 'ol',
              'typ', 'pkg', 'comp', 'step')

words = ('alpha', 'beta', 'gamma', 'delta', 'eta', 'zeta', 'theta',
         'iota', 'kappa', 'mu', 'nu', 'omicron', 'pi', 'rho', 'sigma', 'tau',
         'upsilon', 'omega')

# Can be 'GE', 'LU', 'ADJ', 'LDL', 'CH', 'DM-GE', 'DM-LU', 'DM-charpoly'
# Note, the DM methods require the git version of sympy otherwise
# the fallback method is used.
try:
    from sympy.polys.domainmatrix import DomainMatrix
    matrix_inverse_method = 'DM-charpoly'
except:
    matrix_inverse_method = 'ADJ'
    
matrix_inverse_fallback_method = 'ADJ'

use_units = False

abbreviate_units = True



"""This module provides the FourierDomainExpression class to represent
f-domain (Fourier domain) expressions.

Copyright 2014--2021 Michael Hayes, UCECE

"""

from __future__ import division
from .domains import FourierDomain
from .fourier import inverse_fourier_transform
from .expr import Expr, expr
from .sym import fsym, ssym, tsym, pi
from sympy import Integral, Expr as symExpr

class FourierDomainExpression(FourierDomain, Expr):

    """Fourier domain expression or symbol."""

    var = fsym

    def __init__(self, val, **assumptions):

        check = assumptions.pop('check', True)        
        assumptions['real'] = True
        super(FourierDomainExpression, self).__init__(val, **assumptions)

        expr = self.expr        
        if check and expr.find(ssym) != set() and not expr.has(Integral):
            raise ValueError(
                'f-domain expression %s cannot depend on s' % expr)
        if check and expr.find(tsym) != set() and not expr.has(Integral):
            raise ValueError(
                'f-domain expression %s cannot depend on t' % expr)

    def as_expr(self):
        return FourierDomainExpression(self)

    def angular_fourier(self, **assumptions):
        """Convert to angular Fourier domain."""
        from .symbols import omega
        
        result = self.subs(omega / (2 * pi))
        return result

    def inverse_fourier(self, evaluate=True, **assumptions):
        """Attempt inverse Fourier transform."""

        result = inverse_fourier_transform(self.expr, self.var, tsym, evaluate=evaluate)

        return self.change(result, 'time')

    def IFT(self, **assumptions):
        """Convert to t-domain.   This is an alias for inverse_fourier."""

        return self.inverse_fourier(**assumptions)    
    
    def time(self, **assumptions):
        return self.inverse_fourier(**assumptions)

    def angular_fourier(self, **assumptions):
        """Convert to angular Fourier domain."""
        from .symbols import omega
        
        result = self.subs(omega / (2 * pi))
        return result
    
    def laplace(self, **assumptions):
        """Determine one-side Laplace transform with 0- as the lower limit."""

        result = self.time(**assumptions).laplace()
        return result
    
    def phasor(self, **assumptions):
        """Convert to phasor domain."""

        return self.time(**assumptions).phasor(**assumptions)        

    def plot(self, fvector=None, **kwargs):
        """Plot frequency response at values specified by fvector.  If fvector
        is a tuple, this sets the frequency limits.

        kwargs include:
        axes - the plot axes to use otherwise a new figure is created
        xlabel - the x-axis label
        ylabel - the y-axis label
        ylabel2 - the second y-axis label if needed, say for mag and phase
        xscale - the x-axis scaling, say for plotting as ms
        yscale - the y-axis scaling, say for plotting mV
        plot_type -  'dB_phase', 'mag-phase', 'real-imag', 'mag', 'phase',
        'real', or 'imag'
        in addition to those supported by the matplotlib plot command.
        
        The plot axes are returned.

        There are many plotting options, see lcapy.plot and
        matplotlib.pyplot.plot.

        For example:
            V.plot(fvector, log_frequency=True)
            V.real.plot(fvector, color='black')
            V.phase.plot(fvector, color='black', linestyle='--')

        By default complex data is plotted as separate plots of magnitude (dB)
        and phase.

        """

        from .plot import plot_frequency
        return plot_frequency(self, fvector, **kwargs)


def fexpr(arg, **assumptions):
    """Create FourierDomainExpression object.  If `arg` is fsym return f"""

    if arg is fsym:
        return f
    return FourierDomainExpression(arg, **assumptions)


from .expressionclasses import expressionclasses

classes = expressionclasses.register('fourier', FourierDomainExpression)

f = FourierDomainExpression('f')



from lcapy import *
import unittest


class LcapyTester(unittest.TestCase):
    """Unit tests for lcapy"""

    def assertEqual2(self, ans1, ans2, comment):

        try:
            self.assertEqual(ans1, ans2, comment)
        except AssertionError as e:
            pprint(ans1)
            pprint(ans2)
            raise AssertionError(e)

    def test_phasor(self):

        a = phasor(3, omega=7)
        self.assertEqual(a.omega, 7, 'omega')
        self.assertEqual(a.phase, 0, 'phase')
        self.assertEqual(a.magnitude, 3, 'magnitude')        
        
        a = phasor(-3, omega=7)
        self.assertEqual(a.phase, pi, 'phase')                
        self.assertEqual(a.magnitude, 3, 'magnitude')

        a = phasor(-3 + 4j, omega=7)
        self.assertEqual(a.magnitude, 5, 'magnitude')                        

        v = voltage('cos(5 * t)')
        i = current('cos(5 * t)')
        V = v.as_phasor()
        I = i.as_phasor()

        Z = V / I
        self.assertEqual(Z.is_impedance, True, 'impedance')
        self.assertEqual(Z.omega, 5, 'omega')

    def test_phasor_rms(self):        

        p = phasor(2)
        
        self.assertEqual(p.rms(), sqrt(2), 'rms')
        
        



"""This module provides support for Fourier transforms.  It acts as a
quantity for SymPy's Fourier transform.  It calculates the bilateral
Fourier transform using:

   S(f) = \int_{-\infty}^{\infty} s(t) e^{-j * 2 * \pi * t} dt

It also allows functions that strictly do not have a Fourier transform
by using Dirac deltas.  For example, a, cos(a * t), sin(a * t), exp(j
* a * t).


Copyright 2016--2021 Michael Hayes, UCECE

"""

# TODO:
# Add DiracDelta(t, n)
# Simplify  (-j * DiracDelta(f - 1) +j * DiracDelta(f + 1)).inverse_fourier()
# This should give 2 * sin(2 * pi * t)

import sympy as sym
from .sym import sympify, AppliedUndef, j, pi, symsimplify
from .utils import factor_const, scale_shift
from .extrafunctions import rect, sinc

__all__ = ('FT', 'IFT')


fourier_cache = {}

def fourier_sympy(expr, t, f):

    result = sym.fourier_transform(expr, t, f)
    if expr != 0 and result == 0:
        # There is a bug in SymPy where it returns 0.
        raise ValueError('Could not compute Fourier transform for ' + str(expr))

    if isinstance(result, sym.FourierTransform):
        raise ValueError('Could not compute Fourier transform for ' + str(expr))
    
    return result


def fourier_func(expr, t, f, inverse=False):

    if not isinstance(expr, AppliedUndef):
        raise ValueError('Expecting function for %s' % expr)

    scale, shift = scale_shift(expr.args[0], t)

    fsym = sympify(str(f))
    
    # Convert v(t) to V(f), etc.
    name = expr.func.__name__
    if inverse:
        func = name[0].lower() + name[1:] + '(%s)' % f
    else:
        func = name[0].upper() + name[1:] + '(%s)' % f

    result = sympify(func).subs(fsym, f / scale) / abs(scale)

    if shift != 0:
        if inverse:
            shift = -shift
        result = result * sym.exp(2 * sym.I * sym.pi * f * shift / scale)
    
    return result


def fourier_function(expr, t, f, inverse=False):

    # Handle expressions with a function of FOO, e.g.,
    # v(t), v(t) * y(t),  3 * v(t) / t, v(4 * a * t), etc.,
    
    if not expr.has(AppliedUndef):
        raise ValueError('Could not compute Fourier transform for ' + str(expr))

    const, expr = factor_const(expr, t)
    
    if isinstance(expr, AppliedUndef):
        return fourier_func(expr, t, f, inverse) * const
    
    tsym = sympify(str(t))
    expr = expr.subs(tsym, t)

    rest = sym.S.One
    undefs = []
    for factor in expr.as_ordered_factors():
        if isinstance(factor, AppliedUndef):
            if factor.args[0] != t:
                raise ValueError('Weird function %s not of %s' % (factor, t))
            undefs.append(factor)
        else:
            rest *= factor

    if rest.has(AppliedUndef):
        # Have something like 1/v(t)
        raise ValueError('Cannot compute Fourier transform of %s' % rest)
            
    exprs = undefs
    if rest.has(t):
        exprs = exprs + [rest]
        rest = sym.S.One

    result = fourier_term(exprs[0], t, f, inverse) * rest
        
    if len(exprs) == 1:
        return result * const

    dummy = 'tau' if inverse else 'nu'

    for m in range(len(exprs) - 1):
        if m == 0:
            nu = sympify(dummy)
        else:
            nu = sympify(dummy + '_%d' % m)
        expr2 = fourier_term(exprs[m + 1], t, f, inverse)
        result = sym.Integral(result.subs(f, f - nu) * expr2.subs(f, nu),
                              (nu, -sym.oo, sym.oo))
    
    return result * const

def fourier_term(expr, t, f, inverse=False):

    const, expr = factor_const(expr, t)
    
    if isinstance(expr, AppliedUndef):
        return fourier_func(expr, t, f, inverse) * const
    
    # TODO add u(t) <-->  delta(f) / 2 - j / (2 * pi * f)
    
    if expr.has(AppliedUndef):
        # Handle v(t), v(t) * y(t),  3 * v(t) / t etc.
        return fourier_function(expr, t, f, inverse) * const

    # Check for constant.
    if not expr.has(t):
        return expr * sym.DiracDelta(f) * const

    one = sym.S.One
    const1 = const
    other = one
    exps = one
    factors = expr.expand().as_ordered_factors()    
    for factor in factors:
        if not factor.has(t):
            const1 *= factor
        else:
            if factor.is_Function and factor.func == sym.exp:
                exps *= factor
            else:
                other *= factor

    sf = -f if inverse else f
    
    if other != 1 and exps == 1:
        if other == t:
            return const1 * sym.I / (2 * sym.pi) * sym.DiracDelta(sf, 1)
        elif other == t**2:
            return -const1 / (2 * sym.pi)**2 * sym.DiracDelta(sf, 2)
        # TODO check for other powers of t...
        elif other == sym.sign(t):
            return const1 / (sym.I * sym.pi * sf)
        elif other == sym.sign(t) * t:
            return -const1 * 2 / (2 * sym.pi * f)**2
        elif other == sym.Heaviside(t):
            return const1 / (sym.I * 2 * sym.pi * f) + const1 * sym.DiracDelta(sf) / 2
        elif other == 1 / t:
            return -const1 * sym.I * sym.pi * sym.sign(sf)
        elif other == 1 / t**2:
            return -const1 * 2 * sym.pi**2 * sf * sym.sign(sf)        
        elif other.is_Function and other.func == sym.Heaviside and other.args[0].has(t):
            # TODO, generalise use of similarity and shift theorems for other functions and expressions
            scale, shift = scale_shift(other.args[0], t)
            return (const1 / (sym.I * 2 * sym.pi * sf / scale) / abs(scale) + const1 * sym.DiracDelta(sf) / 2) * sym.exp(sym.I * 2 * sym.pi * sf /scale * shift)
        elif other == sym.Heaviside(t) * t:
            return -const1 / (2 * sym.pi * f)**2 + const1 * sym.I * sym.DiracDelta(sf, 1) / (4 * pi)
        elif other.is_Function and other.func == sinc and other.args[0].has(t):
            scale, shift = scale_shift(other.args[0], t)            
            return const1 * rect(sf / scale) * sym.exp(sym.I * 2 * sym.pi * sf /scale * shift) / abs(scale)
        elif other.is_Function and other.func == rect and other.args[0].has(t):        
            return const1 * sinc(sf / scale) * sym.exp(sym.I * 2 * sym.pi * sf /scale * shift) / abs(scale)

        # Sympy incorrectly gives exp(-a * t) instead of exp(-a * t) *
        # Heaviside(t)
        if other.is_Pow and other.args[1] == -1:
            foo = other.args[0]
            if foo.is_Add and foo.args[1].has(t):
                bar = foo.args[1] / t
                if not bar.has(t) and bar.has(sym.I):
                    a = -(foo.args[0] * 2 * sym.pi * sym.I) / bar
                    return const1 * sym.exp(-a * sf) * sym.Heaviside(sf * sym.sign(a))

        if expr == t * sym.DiracDelta(t, 1):
            return const * sf / (-sym.I * 2 * sym.pi)
                
        # Punt and use SymPy.  Should check for t**n, t**n * exp(-a * t), etc.
        return const * fourier_sympy(expr, t, sf)

    args = exps.args[0]
    foo = args / t
    if foo.has(t):
        # Have exp(a * t**n), SymPy might be able to handle this
        return const * fourier_sympy(expr, t, sf)

    if exps != 1 and foo.has(sym.I):
        return const1 * sym.DiracDelta(sf - foo / (sym.I * 2 * sym.pi))
        
    return const * fourier_sympy(expr, t, sf)


def fourier_transform(expr, t, f, inverse=False, evaluate=True):
    """Compute bilateral Fourier transform of expr.

    Undefined functions such as v(t) are converted to V(f)

    This also handles some expressions that do not really have a
    Fourier transform, such as a, cos(a * t), sin(a * t), exp(I * a *
    t).  These expressions all require the use of the Dirac delta."""

    if expr.is_Equality:
        return sym.Eq(fourier_transform(expr.args[0], t, f, inverse),
                      fourier_transform(expr.args[1], t, f, inverse))

    if not evaluate:
        if inverse:
            result = sym.Integral(expr * sym.exp(j * 2 * pi * f * t), (f, -sym.oo, sym.oo))
        else:
            result = sym.Integral(expr * sym.exp(-j * 2 * pi * f * t), (t, -sym.oo, sym.oo))
        return result

    key = (expr, t, f, inverse)
    if key in fourier_cache:
        return fourier_cache[key]

    if not inverse and expr.has(f):
        raise ValueError('Cannot Fourier transform for expression %s that depends on %s' % (expr, f))


    if expr.is_Piecewise and expr.args[0].args[1].has(t >= 0):
        raise ValueError('Cannot Fourier transform for expression %s that is unknown for t < 0' % expr)
    
    if inverse:
        t, f = f, t

    # The variable may have been created with different attributes,
    # say when using sym.sympify('DiracDelta(t)') since this will
    # default to assuming that t is complex.  So if the symbol has the
    # same representation, convert to the desired one.
    var = sym.Symbol(str(t))
    expr = expr.replace(var, t)

    orig_expr = expr

    # sym.rewrite(sym.exp) can create exp(log...)
    if expr.has(sym.sin):
        expr = expr.replace(lambda expr: expr.is_Function and expr.func == sym.sin,
                            lambda expr: expr.rewrite(sym.exp))
    if expr.has(sym.cos):
        expr = expr.replace(lambda expr: expr.is_Function and expr.func == sym.cos,
                            lambda expr: expr.rewrite(sym.exp))        

    terms = expr.expand().as_ordered_terms()
    result = 0

    try:
        for term in terms:
            result += fourier_term(symsimplify(term), t, f, inverse=inverse)
    except ValueError:
        raise ValueError('Could not compute Fourier transform for ' + str(orig_expr))

    fourier_cache[key] = result
    return result


def inverse_fourier_transform(expr, f, t, evaluate=True):
    """Compute bilateral inverse Fourier transform of expr.

    Undefined functions such as V(f) are converted to v(t)

    This also handles some expressions that do not really have an
    inverse Fourier transform, such as a, cos(a * f), sin(a * f), exp(I *
    a * f)."""
    
    if expr.has(t):
        raise ValueError('Cannot inverse Fourier transform for expression %s that depends on %s' % (expr, t))
    
    result = fourier_transform(expr, t, f, inverse=True, evaluate=evaluate)
    return symsimplify(result)


def FT(expr, t, f, **assumptions):
    """Compute bilateral Fourier transform of expr.

    Undefined functions such as v(t) are converted to V(f)

    This also handles some expressions that do not really have a Fourier
    transform, such as a, cos(a * t), sin(a * t), exp(I * a * t)."""    
    
    return fourier_transform(expr, t, f, **assumptions)


def IFT(expr, f, t, **assumptions):
    """Compute bilateral inverse Fourier transform of expr.

    Undefined functions such as V(f) are converted to v(t)

    This also handles some expressions that do not really have an
    inverse Fourier transform, such as a, cos(a * f), sin(a * f), exp(I *
    a * f)."""    
    
    return inverse_fourier_transform(expr, f, t, **assumptions)



"""This module provides voltage support.

Copyright 2020 Michael Hayes, UCECE

"""
from .expr import expr
from .units import u as uu
from .classmap import domain_kind_to_symbol, domain_kind_quantity_to_class


def Vname(name, kind, cache=False):

    # Not caching is a hack to avoid conflicts of Vn1 with Vn1(s) etc.
    # when using subnetlists.  The alternative is a proper context
    # switch.  This would require every method to set the context.

    if kind == 't':
        name = name.lower()
        
    undef = domain_kind_to_symbol(kind, name)
    cls = domain_kind_quantity_to_class(kind, 'voltage')
    return cls(undef, cache=cache)


def Vtype(kind):

    return domain_kind_quantity_to_class(kind, 'voltage')    


def voltage(arg, **assumptions):

    expr1 = expr(arg, **assumptions)

    if 'nid' in assumptions:
        from .noisefexpr import FourierNoiseDomainVoltage
        from .noiseomegaexpr import AngularFourierNoiseDomainVoltage        
        
        if expr1.is_fourier_domain or expr1.is_constant_domain:
            expr1 = FourierNoiseDomainVoltage(expr1)
        elif expr1.is_angular_fourier_domain:
            expr1 = AngularFourierNoiseDomainVoltage(expr1)
        else:
            raise ValueError('Cannot represent noise voltage in %s domain' % expr1.domain)
                
    try:
        expr1 = expr1.as_voltage()
    except:        
        raise ValueError('Cannot represent %s(%s) as voltage' % (expr1.__class__.__name__, expr1))
    
    return expr1.apply_unit(uu.volts)


def noisevoltage(arg, **assumptions):
    """Create a new noise voltage with specified amplitude spectral density."""

    nid = assumptions.get('nid', None)
    positive = assumptions.get('positive', True)    

    return voltage(arg, nid=None, positive=positive, **assumptions)


def phasorvoltage(arg, omega=None, **assumptions):
    from .phasor import phasor
    
    return phasor(arg, omega, **assumptions).as_voltage()



"""This module provides the Quantity class, for tracking of an expression quantity.

Copyright 2020--2021 Michael Hayes, UCECE

"""

class Quantity(object):

    quantity = ''    
    is_power = False
    is_voltage = False
    is_current = False
    is_impedance = False
    is_admittance = False
    is_transfer = False
    is_immitance = False
    is_always_causal = False
    is_undefined = False
    is_voltagesquared = False
    is_currentsquared = False
    is_impedancesquared = False
    is_admittancesquared = False
    part = ''


class UndefinedQuantity(Quantity):

    quantity = 'undefined'        
    is_undefined = True    




"""This module provides the DiscreteFourierDomainExpression class to
 represent k-domain (discrete Fourier domain) expressions.

Copyright 2020--2021 Michael Hayes, UCECE

"""

from __future__ import division
from .domains import DiscreteFourierDomain
from .fourier import inverse_fourier_transform
from .functions import exp
from .sym import j, oo, pi
from .seqexpr import SequenceExpression
from .dsym import nsym, ksym, zsym
from .dft import IDFT
from sympy import Sum

__all__ = ('kexpr', )

class DiscreteFourierDomainExpression(DiscreteFourierDomain, SequenceExpression):
    """Discrete Fourier domain expression or symbol."""

    var = ksym

    def __init__(self, val, **assumptions):

        check = assumptions.pop('check', True)
        assumptions['real'] = True
        
        super(DiscreteFourierDomainExpression, self).__init__(val, **assumptions)

        expr = self.expr
        if check and expr.find(zsym) != set():
            raise ValueError(
                'k-domain expression %s cannot depend on z' % expr)
        if check and expr.find(nsym) != set() and not expr.has(Sum):
            raise ValueError(
                'k-domain expression %s cannot depend on n' % expr)

    def as_expr(self):
        return DiscreteFourierDomainExpression(self)

    def plot(self, kvector=None, **kwargs):
        """Plot frequency response at values specified by kvector.  If kvector
        is a tuple, this sets the frequency limits.

        kwargs include:
        axes - the plot axes to use otherwise a new figure is created
        xlabel - the x-axis label
        ylabel - the y-axis label
        ylabel2 - the second y-axis label if needed, say for mag and phase
        xscale - the x-axis scaling, say for plotting as ms
        yscale - the y-axis scaling, say for plotting mV
        plot_type -  'dB_phase', 'mag-phase', 'real-imag', 'mag', 'phase',
        'real', or 'imag'
        in addition to those supported by the matplotlib plot command.
        
        The plot axes are returned.

        There are many plotting options, see lcapy.plot and
        matplotlib.pyplot.plot.

        For example:
            V.plot(kvector, log_frequency=True)
            V.real.plot(kvector, color='black')
            V.phase.plot(kvector, color='black', linestyle='--')

        By default complex data is plotted as separate plots of magnitude (dB)
        and phase.

        """

        from .plot import plot_frequency
        return plot_frequency(self, kvector, **kwargs)

    def IDFT(self, N=None, evaluate=True):

        from .nexpr import n

        if N is None:
            from .sym import sympify
            
            N = sympify('N')

        result = IDFT(self.expr, ksym, nsym, N, evaluate=evaluate)
        return self.change(result, domain='discrete time')
    
    def ZT(self, **assumptions):
        return self.IDFT().ZT(**assumptions)

    
def kexpr(arg, **assumptions):
    """Create kExpr object.  If `arg` is ksym return k"""

    if arg is ksym:
        return k
    return DiscreteFourierDomainExpression(arg, **assumptions)


from .expressionclasses import expressionclasses

classs = expressionclasses.register('discrete fourier', DiscreteFourierDomainExpression)

k = DiscreteFourierDomainExpression('k')



from lcapy import *
from lcapy.fexpr import FourierDomainExpression
from lcapy.expr import TimeDomainExpression
import unittest


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_fourier(self):

        self.assertEqual(DiracDelta(t).fourier(), 1, "DiracDelta(t)")
        self.assertEqual(TimeDomainExpression('x(t)').fourier(), FourierDomainExpression('X(f)'), "x(t)")
        self.assertEqual(TimeDomainExpression('5 * x(t)').fourier(), FourierDomainExpression('5 * X(f)'), "5 * x(t)")

    def test_inverse_fourier(self):

        self.assertEqual((f * 0 + 1).inverse_fourier(), DiracDelta(t),
                         "1")
        self.assertEqual((f * 0 + 10).inverse_fourier(), 10
                         * DiracDelta(t), "0")
        self.assertEqual(FourierDomainExpression('V(f)').inverse_fourier(), TimeDomainExpression('v(t)'), "V(f)")
        self.assertEqual(FourierDomainExpression('10 * V(f)').inverse_fourier(),
                         TimeDomainExpression('10 * v(t)'), "V(f)")
        self.assertEqual(TimeDomainExpression('v(t)').fourier().inverse_fourier(),
                         TimeDomainExpression('v(t)'), "v(t)")
        self.assertEqual(TimeDomainExpression('v(t/2)').fourier().inverse_fourier(),
                         TimeDomainExpression('v(t/2)'), "v(t/2)")        
                         
    def test_fourier2(self):

        # 1
        self.assertEqual((t * 0 + 1).fourier(), DiracDelta(f))
        self.assertEqual((t * 0 + 1).fourier().inverse_fourier(), 1)

        # t
        self.assertEqual(t.fourier(), j * DiracDelta(f, 1) / (2 * pi))
        self.assertEqual(t.fourier().inverse_fourier(), t)

        # t**2
        self.assertEqual((t**2).fourier(), -DiracDelta(f, 2) / (2 * pi)**2)
        self.assertEqual((t**2).fourier().inverse_fourier(), t**2)
        
        # 2 * cos(2 * pi * t)
        self.assertEqual(2 * cos(2 * pi * t).fourier(),
                         DiracDelta(f - 1) + DiracDelta(f + 1))

        # 2 * sin(2 * pi * t)        
        self.assertEqual(2 * sin(2 * pi * t).fourier(),
                         -j * DiracDelta(f - 1) + j * DiracDelta(f + 1))

        # exp(j * 2 * pi * t)
        self.assertEqual(exp(j * 2 * pi * t).fourier(), DiracDelta(f - 1))
        self.assertEqual(exp(j * 2 * pi * t).fourier().inverse_fourier(),
                         exp(j * 2 * pi * t))

        # exp(-j * 2 * pi * t)        
        self.assertEqual(exp(-j * 2 * pi * t).fourier(), DiracDelta(f + 1))
        self.assertEqual(exp(-j * 2 * pi * t).fourier().inverse_fourier(),
                         exp(-j * 2 * pi * t))

        # u(t)
        self.assertEqual(u(t).fourier(), DiracDelta(f) / 2 + 1 / (j * 2 * pi * f))
        # TODO teach SymPy that sign(t) + 1 = 2 * u(t)
        #self.assertEqual(u(t).fourier().inverse_fourier(), u(t))

        # t * u(t)
        self.assertEqual((t * u(t)).fourier(), -1 / (2 * pi * f)**2 + j * DiracDelta(f, 1) / (4 * pi))
        # TODO teach SymPy that sign(t) + 1 = 2 * u(t)        
        #self.assertEqual((t * u(t)).fourier().inverse_fourier(), t * u(t))        

        # sign(t)
        self.assertEqual(sign(t).fourier(), 1 / (j * pi * f))
        self.assertEqual(sign(t).fourier().inverse_fourier(), sign(t))

        # t * sign(t)
        self.assertEqual((t * sign(t)).fourier(), -1 / (2 * pi**2 * f**2))
        self.assertEqual((t * sign(t)).fourier().inverse_fourier(), t * sign(t))

        # 2 * cos(2 * pi * t + 3)
        self.assertEqual(2 * cos(2 * pi * t + 3).fourier(),
                         DiracDelta(f - 1) * exp(3 * j) +
                         DiracDelta(f + 1) * exp(-3 * j))
        



from .quantity import Quantity

class CurrentMixin(Quantity):

    quantity = 'current'
    quantity_label = 'Current'
    units = 'A'
    is_current = True

    def cpt(self):
        from .oneport import I, Idc, Iac
        
        if self.is_dc:
            return Idc(self.expr)
        elif self.is_phasor_domain:        
            return Iac(self, omega=self.omega)            
        
        return I(self.expr)

    
    

    
    



from lcapy import *
from lcapy.polyphase import *
from lcapy.discretetime import *
import unittest


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_polyphase(self):

        A = symbol('A')
        alpha = polyphase_alpha(3)
        
        V = PhaseVoltageVector((A, A * alpha, A * alpha**2))



from lcapy import *

a = Circuit('VRC2.sch')

H = a.P1.transfer('P2')

v_i = voltage(sin(3 * t) * u(t))
V_i = v_i(s)
V_o = H * V_i
v_o = V_o(t)

ax = v_i.plot((-1, 10), label='input')
ax = v_o.plot((-1, 10), axes=ax, label='output')
ax.legend()

from matplotlib.pyplot import savefig
savefig(__file__.replace('.py', '.png'), bbox_inches='tight')








"""
Copyright 2014--2020 Michael Hayes, UCECE
"""

from __future__ import division
from .expr import expr
from .sexpr import s
from .printing import latex, pretty
from .schematic import Schematic
from .circuit import Circuit
from .state import state

class Network(object):
    """This is the base class for network objects."""

    is_voltage_source = False
    is_current_source = False
    is_inductor = False
    is_capacitor = False
    is_resistor = False
    is_conductor = False    

    # True if initial conditions are zero (or unspecified).
    zeroic = True

    # None if component does not have initial conditions.
    # True if initial conditions are specified.
    # False if initial conditions are not specified.
    has_ic = None

    netname = ''
    netkeyword = ''

    def _tweak_args(self):

        if not hasattr(self, 'args'):
            return ()

        modargs = []
        for arg in self.args:
            arg = expr(arg)
            modargs.append(arg)
        return modargs

    def __repr__(self):

        argsrepr = ', '.join([repr(arg) for arg in self.args])
        return '%s(%s)' % (self.__class__.__name__, argsrepr)

    def __str__(self):

        argsrepr = ', '.join([repr(arg) for arg in self.args])
        return '%s(%s)' % (self.__class__.__name__, argsrepr)

    def _repr_pretty_(self, p, cycle):

        p.text(self.pretty())

    def _repr_latex_(self):

        return '$%s$' % self.latex()

    def pretty(self, **kwargs):

        argsrepr = ', '.join([pretty(arg, **kwargs) for arg in self._tweak_args()])
        return '%s(%s)' % (self.__class__.__name__, argsrepr)

    def latex(self, **kwargs):

        argsrepr = ', '.join([latex(arg, **kwargs) for arg in self._tweak_args()])
        return '\\mathrm{%s}(%s)' % (self.__class__.__name__, argsrepr)

    def pprint(self):

        print(self.pretty())
    
    @property
    def analysis(self):
        return self.cct.analysis

    def describe(self):
        """Print a message describing how network is solved."""
        return self.cct.describe()
    
    def simplify(self):

        return self

    def _add_elements(self):

        netlist = self.netlist()
        for net in netlist.split('\n'):
            self._add(net)

        # Hack, create ground reference.
        self._add('W %d 0' % (self._node - 1))

    def _net_make(self, netlist, n1=None, n2=None, dir='right'):

        net = self
        if n2 == None:
            n2 = net._node
        if n1 == None:
            n1 = net._node
        
        netname = net.__class__.__name__ if net.netname == '' else net.netname

        netid = netlist._make_id(netname)
        if net.netkeyword != '':
            return '%s%s %s %s %s %s; %s' % (netname, netid,
                                             n1, n2, 
                                             net.netkeyword,
                                             netlist._netargs(net), dir)
        else:
            return '%s%s %s %s %s; %s' % (netname, netid,
                                          n1, n2, netlist._netargs(net), dir)

    @property 
    def _depths(self):
        return [net._depth for net in self.args]
        
    @property 
    def _depth(self):
        from .oneport import Ser, Par
        
        if not isinstance(self, (Ser, Par)):
            return 0
        
        depths = self._depths
        return 1 + max(depths)
        
    def netlist(self, form='horizontal', evalf=None):
        """Create a netlist.

        `form` can be 'horizontal', 'vertical', or 'ladder'.

        `evalf` can be False or an integer specifying the number of
        decimal places used to evaluate floats.
        """

        if form == 'ladder':
            from .laddermaker import LadderMaker
            return LadderMaker(self, form=form, evalf=evalf)()

        from .netlistmaker import NetlistMaker        
        return NetlistMaker(self, form=form, evalf=evalf)()

    def pdb(self):
        """Enter the python debugger."""
        
        import pdb; pdb.set_trace()
        return self
    
    def sch(self, form, evalf=False):
        """Convert a Network object into a Schematic object."""

        netlist = self.netlist(form=form, evalf=evalf)
        sch = Schematic()
        for net in netlist.split('\n'):
            sch.add(net)

        return sch

    def draw(self, filename=None, form='horizontal', evalf=False, **kwargs):
        """Draw schematic of network.

        filename specifies the name of the file to produce.  If None,
        the schematic is displayed on the screen.

        Note, if using Jupyter, then need to first issue command %matplotlib inline

        `form` is either 'horizontal', 'vertical', or 'ladder'.

        `evalf` can be False or an integer specifying the number of
        decimal places used to evaluate floats.

        `kwargs` include:
           label_ids: True to show component ids
           label_values: True to display component values
           draw_nodes: True to show all nodes, False to show no nodes, 
             'primary' to show primary nodes,
             'connections' to show nodes that connect more than two components,
             'all' to show all nodes
           label_nodes: True to label all nodes, False to label no nodes, 
             'primary' to label primary nodes,
             'alpha' to label nodes starting with a letter,
             'pins' to label nodes that are pins on a chip,
             'all' to label all nodes
           style: 'american', 'british', or 'european'
           scale: schematic scale factor, default 1.0
           node_spacing: spacing between component nodes, default 2.0
           cpt_size: size of a component, default 1.5
           dpi: dots per inch for png files
           help_lines: distance between lines in grid, default 0.0 (disabled)
           debug: True to display debug information
        """

        if 'label_ids' not in kwargs:
            kwargs['label_ids'] = False
        if 'label_values' not in kwargs:
            kwargs['label_values'] = True
        if 'label_nodes' not in kwargs:
            kwargs['label_nodes'] = False
        if 'draw_nodes' not in kwargs:
            kwargs['draw_nodes'] = 'connections'
        
        self.sch(form=form, evalf=evalf).draw(filename=filename, **kwargs)
        
    @property
    def cct(self):
        """Convert a Network object into a Circuit object."""

        if hasattr(self, '_cct'):
            return self._cct

        netlist = self.netlist()
        cct = Circuit()
        for net in netlist.split('\n'):
            cct.add(net)
        self._cct = cct
        return cct

    def circuit(self):
        """Convert a Network object into a Circuit object."""

        return self.cct

    @property    
    def initial_value_problem(self):
        return self.cct.initial_value_problem

    @property    
    def is_ivp(self):
        return self.cct.is_ivp

    @property
    def is_dc(self):
        return self.cct.is_dc

    @property    
    def is_ac(self):
        return self.cct.is_ac

    @property    
    def is_causal(self):
        return self.cct.is_causal

    @property
    def has_dc(self):
        return self.cct.has_dc

    @property    
    def has_ac(self):
        return self.cct.has_ac

    @property    
    def has_transient(self):
        return self.cct.has_transient

    @property    
    def kinds(self):
        """Return list of transform domain kinds."""        
        return self.cct.kinds

    @property
    def symbols(self):
        """Return dictionary of symbols defined in the network."""
        
        return self.cct.symbols

    @property
    def all_symbols(self):
        """Return dictionary of symbols defined in the network and global
        symbols."""

        symbols = self.symbols
        symbols.update(state.global_context.symbols)
        return symbols

    def transform(self, form='causerI'):
        """Transform the network into an alternative form.  The transformation
        is performed using network synthesis of the network's
        impedance (note, this ignores the sources).  `form` includes:
        cauerI, cauerII, fosterI, fosterII.

        Note some methods generate networks with negative value
        components."""

        return self.Z(s).network(form)



"""This module provides the NetlistNamespace class.

Copyright 2020 Michael Hayes, UCECE

"""

from .schematic import Schematic

class NetlistNamespace(object):
    """This class allows elements, nodes, or other namespaces
    in a heirachical namespace to be accessed by name, via __getattr__.
    
    For example,

    >>> a = Circuit('''
    ... b.L1 1 2'''
    >>> a.b.L1.v

    """

    def __init__(self, namespace, netlist):

        # The namespace name, e.g., a or a.b
        self.namespace = namespace
        self._netlist = netlist
        # The children namespaces
        self.namespaces = {}
        self.allow_anon = False

    def __getitem__(self, name):
        """Return element or node by name."""

        netlist = self._netlist
        
        # If name is an integer, convert to a string.
        if isinstance(name, int):
            name = '%d' % name

        name = self.namespace + '.' + name
            
        if name in netlist.nodes:
            return netlist.nodes[name]

        if name in netlist._elements:
            return netlist._elements[name]

        # Try first anonymous name.
        if name + 'anon1' in netlist._elements:
            return netlist._elements[name + 'anon1']

        if name in self.namespaces:
            return self.namespaces[name]
        
        raise AttributeError('Unknown element or node name %s' % name)

    def __getattr__(self, attr):
        """Return element, node, or another NetlistNamespace object by name.
        This gets called if there is no explicit attribute attr for
        this instance.  This is primarily for accessing elements and
        non-numerical node names.  It also gets called if the called
        attr throws an AttributeError exception.  The annoying thing
        is that hasattr uses getattr and checks for an exception.

        """

        return self.__getitem__(attr)

    @property
    def components(self):
        """Return ordered dictionary of components."""

        return self.elements

    def netlist(self):
        """Return the current netlist for this namespace."""

        nlist = self._netlist
        
        return '\n'.join([str(cpt) for cpt in nlist._elements.values() if str(cpt).startswith(self.namespace)])

    def __repr__(self):
        
        return self.netlist()

    @property
    def sch(self):
        """Generate schematic for this namespace."""        

        if hasattr(self, '_sch'):
            return self._sch

        sch = Schematic(allow_anon=self.allow_anon)

        netlist = self._netlist.netlist()
        for net in netlist.split('\n'):
            if net.startswith(self.namespace):
                sch.add(net)

        self._sch = sch
        return sch

    def draw(self, filename=None, **kwargs):
        """Draw schematic of subnetlist.

        filename specifies the name of the file to produce.  If None,
        the schematic is displayed on the screen.
        
        Note, if using Jupyter, then need to first issue command %matplotlib inline

        kwargs include:
           label_ids: True to show component ids
           label_values: True to display component values
           draw_nodes:
              True to show all nodes,
              False or 'none' to show no nodes, 
              'primary' to show primary nodes,
              'connections' to show nodes that connect more than two components,
              'all' to show all nodes
           label_nodes:
               True to label all nodes,
               False or 'none' to label no nodes, 
               'primary' to label primary nodes,
               'alpha' to label nodes starting with a letter,
               'pins' to label nodes that are pins on a chip,
               'all' to label all nodes
           style: 'american', 'british', or 'european'
           scale: schematic scale factor, default 1.0
           node_spacing: spacing between component nodes, default 2.0
           cpt_size: size of a component, default 1.5
           dpi: dots per inch for png files
           help_lines: distance between lines in grid, default 0.0 (disabled)
           debug: True to display debug information
        """

        cct = self
        if kwargs.pop('s_model', False):
            cct = cct._s_model()

        return cct.sch.draw(filename=filename, **kwargs)




"""This module provides support for discrete Fourier transforms. 

 It calculates the discrete Fourier transform using:

   X(k) = \sum_{n=0}^{N-1} x(n) e^{-j * 2 * \pi * n * k / N}

Copyright 2020 Michael Hayes, UCECE

"""

import sympy as sym
from .sym import sympify, AppliedUndef, j, pi
from .functions import UnitImpulse, UnitStep, exp
from .utils import factor_const, scale_shift
from .matrix import Matrix

__all__ = ('DFT', 'IDFT', 'DFTmatrix', 'IDFTmatrix')


discrete_fourier_cache = {}


def discrete_fourier_sympy(expr, n, k, N):

    foo = expr * sym.exp(-2 * j * pi * n * k / N)
    result = sym.summation(foo, (n, 0, N - 1))
    
    return result


def discrete_fourier_func(expr, n, k, N, inverse=False):

    if not isinstance(expr, AppliedUndef):
        raise ValueError('Expecting function for %s' % expr)

    scale, shift = scale_shift(expr.args[0], n)

    fsym = sympify(str(k))
    
    # Convert v(n) to V(k), etc.
    name = expr.func.__name__
    if inverse:
        func = name[0].lower() + name[1:] + '(%s)' % k
    else:
        func = name[0].upper() + name[1:] + '(%s)' % k

    result = sympify(func).subs(fsym, k / scale) / abs(scale)

    if shift != 0:
        if inverse:
            shift = -shift
        result = result * sym.exp(2 * sym.I * sym.pi * k * shift / scale)
    
    return result


def discrete_fourier_function(expr, n, k, N, inverse=False):

    # Handle expressions with a function of FOO, e.g.,
    # v(n), v(n) * y(n),  3 * v(n) / n, v(4 * a * n), etc.,
    
    if not expr.has(AppliedUndef):
        raise ValueError('Could not compute Fourier transform for ' + str(expr))

    const, expr = factor_const(expr, n)
    
    if isinstance(expr, AppliedUndef):
        return discrete_fourier_func(expr, n, k, N, inverse) * const
    
    tsym = sympify(str(n))
    expr = expr.subs(tsym, n)

    rest = sym.S.One
    undefs = []
    for factor in expr.as_ordered_factors():
        if isinstance(factor, AppliedUndef):
            if factor.args[0] != n:
                raise ValueError('Weird function %s not of %s' % (factor, n))
            undefs.append(factor)
        else:
            rest *= factor

    if rest.has(AppliedUndef):
        # Have something like 1/v(n)
        raise ValueError('Cannot compute Fourier transform of %s' % rest)
            
    exprs = undefs
    if rest.has(n):
        exprs = exprs + [rest]
        rest = sym.S.One

    result = discrete_fourier_term(exprs[0], n, k, N, inverse) * rest
        
    if len(exprs) == 1:
        return result * const

    dummy = 'm' if inverse else 'l'

    for m in range(len(exprs) - 1):
        if m == 0:
            nu = sympify(dummy)
        else:
            nu = sympify(dummy + '_%d' % m)
        expr2 = discrete_fourier_term(exprs[m + 1], n, k, N, inverse)
        # Should be a circular convolution.
        result = sym.Sum(result.subs(k, k - nu) * expr2.subs(k, nu),
                         (nu, 0, N - 1)) / N
    
    return result * const

def discrete_fourier_term(expr, n, k, N, inverse=False):

    const, expr = factor_const(expr, n)

    # Check for constant.
    if not expr.has(n):
        return expr * N * UnitImpulse(k) * const

    if expr.is_Function and expr.func == UnitStep and expr.args[0] == n:
        return const * N * UnitImpulse(k)    

    if expr.is_Function and expr.func == UnitImpulse and expr.has(n):
        scale, shift = scale_shift(expr.args[0], n)
        # This implies a circular shift rather than a delay/advance.
        return const * sym.exp(2 * j * pi * shift * k / N)

    if expr.has(AppliedUndef):
        # Handle v(n), v(n) * y(n), 3 * v(n) / n etc.
        return discrete_fourier_function(expr, n, k, N, inverse) * const

    if expr.is_Function and expr.func == sym.exp and expr.args[0].has(n):
        scale, shift = scale_shift(expr.args[0], n)
        l = scale * N / (j * 2 * pi)
        # TODO, need to wrap the frequency shift, l
        return const * N * sym.exp(2 * j * pi * shift * k / N) * UnitImpulse(k + l)
    
    return const * discrete_fourier_sympy(expr, n, k, N)


def discrete_fourier_transform(expr, n, k, N=None, inverse=False, evaluate=True):
    """Compute bilateral discrete Fourier transform of expr.

    Undefined functions such as x(n) are converted to X(k)
    """

    if expr.is_Equality:
        return sym.Eq(discrete_fourier_transform(expr.args[0], n, k, N, inverse),
                      discrete_fourier_transform(expr.args[1], n, k, N, inverse))    

    if not evaluate:
        foo = expr * sym.exp(-2 * j * pi * n * k / N)
        result = sym.Sum(foo, (n, 0, N - 1))
        if inverse:
            result /= N
        return result
    
    key = (expr, n, k, N, inverse)
    if key in discrete_fourier_cache:
        return discrete_fourier_cache[key]

    if not inverse and expr.has(k):
        raise ValueError('Cannot discrete Fourier transform for expression %s that depends on %s' % (expr, k))


    if expr.is_Piecewise and expr.args[0].args[1].has(n >= 0):
        raise ValueError('Cannot discrete Fourier transform for expression %s that is unknown for n < 0' % expr)
    
    if inverse:
        n, k = k, n

    var = sym.Symbol(str(n))
    expr = expr.replace(var, n)

    orig_expr = expr

    if expr.has(sym.cos) or expr.has(sym.sin):
        expr = expr.rewrite(sym.exp)
        
    terms = expr.expand().as_ordered_terms()
    result = 0

    try:
        for term in terms:
            result += discrete_fourier_term(term, n, k, N, inverse=inverse)
    except ValueError:
        raise ValueError('Could not compute discrete Fourier transform for ' + str(orig_expr))

    if inverse:
        result /= N
    
    discrete_fourier_cache[key] = result
    return result


def inverse_discrete_fourier_transform(expr, k, n, N=None, evaluate=False):
    """Compute bilateral inverse discrete Fourier transform of expr.

    Undefined functions such as X(k) are converted to x(n)
    """
    
    if expr.has(n):
        raise ValueError('Cannot inverse discrete Fourier transform for expression %s that depends on %s' % (expr, n))
    
    result = discrete_fourier_transform(expr, n, k, N, inverse=True, evaluate=evaluate) 
    return sym.simplify(result)


def DFT(expr, n, k, N, **assumptions):
    """Compute bilateral discrete Fourier transform of expr.

    Undefined functions such as x(n) are converted to X(k)
    """
    
    return discrete_fourier_transform(expr, n, k, N, **assumptions)


def IDFT(expr, k, n, N, **assumptions):
    """Compute bilateral inverse discrete Fourier transform of expr.

    Undefined functions such as X(k) are converted to x(n)
    """    
    
    return inverse_discrete_fourier_transform(expr, k, n, N, **assumptions)


def DFTmatrix(N):
    """Return DFT matrix of size `N` x `N`."""    

    w = exp(-j * 2 * pi / N)

    a = Matrix.zeros(N)
    
    for row in range(N):
        for col in range(N):
            a[row, col] = w ** (row * col)
    return a


def IDFTmatrix(N):
    """Return inverse DFT matrix of size `N` x `N`."""

    w = exp(j * 2 * pi / N)

    a = Matrix.zeros(N)
    
    for row in range(N):
        for col in range(N):
            a[row, col] = w ** (row * col)
    return a / N



from lcapy import *
from lcapy.texpr import TimeDomainVoltage, TimeDomainCurrent
from lcapy.noiseomegaexpr import AngularFourierNoiseDomainVoltage, AngularFourierNoiseDomainCurrent
import unittest
import sympy as sym


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def assertEqual2(self, ans1, ans2, comment):

        ans1 = ans1.canonical()
        ans2 = ans2.canonical()

        try:
            self.assertEqual(ans1, ans2, comment)
        except AssertionError as e:
            ans1.pprint()
            ans2.pprint()
            raise AssertionError(e)

    def test_VR_ac1(self):
        """Lcapy: check VR ac network

        """

        a = Vac(4) | R(2)

        self.assertEqual(a.is_dc, False, "DC incorrect")
        self.assertEqual(a.is_ac, True, "AC incorrect")

    def test_VR_dc1(self):
        """Lcapy: check VR dc network

        """

        a = Vdc(4) | R(2)

        self.assertEqual(a.is_ivp, False, "is_ivp incorrect")
        self.assertEqual(a.is_ac, False, "AC incorrect")
        self.assertEqual(a.is_dc, True, "DC incorrect")

    def test_VC_dc1(self):
        """Lcapy: check VC dc network

        """

        a = Vdc(4) + C(2)

        self.assertEqual(a.is_ivp, False, "is_ivp incorrect")
        self.assertEqual(a.is_ac, False, "AC incorrect")
        self.assertEqual(a.is_dc, True, "DC incorrect")

    def test_VC_dc2(self):
        """Lcapy: check VC dc network

        """

        a = Vdc(4) + C(2, 0)

        self.assertEqual(a.is_ivp, True, "is_ivp incorrect")
        self.assertEqual(a.is_ac, False, "AC incorrect")
        self.assertEqual(a.is_dc, False, "DC incorrect")

    def test_thevenin_ac(self):
        """Lcapy: check ac Thevenin conversion

        """
        a = (Vac('1') + C(2)) | R(3)
        
        self.assertEqual(a.norton().isc, TimeDomainCurrent(-2 * omega0 * sin(omega0 * t)),
                         "Isc incorrect")
        self.assertEqual(a.thevenin().isc, TimeDomainCurrent(-2 * omega0 * sin(omega0 * t)),
                         "Isc incorrect")
        
    def test_superposition(self):
        """Lcapy: check network superposition"""

        a = Vac(40) + Vnoise(20) + Vstep(10) + R(5)
        self.assertEqual(a.Voc.dc, 0, "Voc.dc error")
        self.assertEqual(a.Voc.has_dc, False, "Voc.has_dc error")
        self.assertEqual(a.Voc.is_dc, False, "Voc.is_dc error")
        self.assertEqual(a.Voc.has_ac, True, "Voc.has_ac error")
        self.assertEqual(a.Voc.is_ac, False, "Voc.is_ac error")                                
        self.assertEqual2(a.Voc.s, 10 / s, "Voc.s error")
        self.assertEqual(a.Voc.n.expr, AngularFourierNoiseDomainVoltage(20).expr, "Voc.n error")
        self.assertEqual2(a.Isc.s, 2 / s, "Isc.s error")
        # FIXME, this intermittently fails.
        self.assertEqual(a.Isc.n.expr, AngularFourierNoiseDomainCurrent(4).expr, "Isc.n error")
        
    def test_ivp(self):
        """Lcapy: check network with initial values"""

        a = Vstep(10) + C('C1', 5)
        self.assertEqual(a.is_ivp, True, "IVP fail")
        self.assertEqual2(a.Voc.s, 15 / s, "Voc fail")

    def test_causal(self):
        """Lcapy: check network is causal"""

        a = Vstep(10) + C('C1')
        self.assertEqual(a.is_causal, True, "causal fail")
        self.assertEqual(a.Isc.is_causal, True, "causal fail")        
        
    def test_YZ(self):

        a = R(1) + V(2) + R(3)
        self.assertEqual(a.Z, 4, "series Z")
        self.assertEqual(a.Y, expr('1 / 4'), "series Y")

        b = (R(1) + V(2)) | R(3)
        self.assertEqual(b.Z, expr('3 / 4'), "par Z")
        self.assertEqual(b.Y, expr('4 / 3'), "par Y")                        

    def test_attrs(self):

        self.assertEqual(R(1).is_resistor, True, "is_resistor")
        self.assertEqual(G(1).is_conductor, True, "is_conductor")
        self.assertEqual(C(1).is_capacitor, True, "is_capacitor")
        self.assertEqual(L(1).is_inductor, True, "is_inductor")
        self.assertEqual(I(1).is_current_source, True, "is_current_source")
        self.assertEqual(V(1).is_voltage_source, True, "is_voltage_source")                                        



"""This module defines the domain classes.

Copyright 2021 Michael Hayes, UCECE

"""


class Domain(object):
    is_undefined_domain = False
    is_constant_domain = False
    is_time_domain = False
    is_laplace_domain = False
    is_fourier_domain = False
    is_angular_fourier_domain = False
    is_phasor_domain = False
    is_phasor_time_domain = False
    is_phasor_frequency_domain = False
    is_fourier_noise_domain = False
    is_angular_fourier_noise_domain = False    
    is_discrete_time_domain = False
    is_discrete_fourier_domain = False
    is_Z_domain = False
    is_transform_domain = False
    is_superposition_domain = False
    is_one_sided = False


class UndefinedDomain(Domain):
    domain = 'undefined'
    domain_label = 'Undefined'
    domain_units = ''
    is_undefined_domain = True


class ConstantDomain(Domain):    
    domain = 'constant'
    domain_label = 'Constant'    
    domain_units = ''
    is_constant_domain = True
    

class TimeDomain(Domain):
    domain = 'time'
    domain_label = 'Time'
    domain_units = 's'
    is_time_domain = True

    
class LaplaceDomain(Domain):
    domain = 'laplace'        
    domain_label = 'Laplace frequency'
    domain_units = 'rad/s'
    is_laplace_domain = True
    is_transform_domain = True

    
class FourierDomain(Domain):
    domain = 'fourier'    
    domain_label = 'Frequency'
    domain_units = 'Hz'
    is_fourier_domain = True
    is_transform_domain = True    

    
class AngularFourierDomain(Domain):
    domain = 'angular fourier'
    domain_label = 'Angular frequency'
    domain_units = 'rad/s'
    is_angular_fourier_domain = True
    is_transform_domain = True    
    

class PhasorTimeDomain(Domain):
    domain = 'phasor'
    domain_label = ''
    domain_units = ''
    is_phasor_domain = True
    is_transform_domain = True

    
class PhasorFrequencyDomain(Domain):
    domain = 'phasor'
    domain_label = 'Angular Frequency'
    domain_units = 'rad/s'
    is_phasor_domain = True
    is_transform_domain = True            


class FourierNoiseDomain(Domain):
    domain = 'fourier noise'    
    domain_label = 'Frequency'
    domain_units = 'sqrt(Hz)'
    is_fourier_noise_domain = True
    is_transform_domain = True
    is_one_sided = True


class AngularFourierNoiseDomain(Domain):
    domain = 'angular fourier noise'    
    domain_label = 'Angular frequency'
    domain_units = 'sqrt(rad/s)'
    is_angular_fourier_noise_domain = True
    is_transform_domain = True
    is_one_sided = True    
    

class DiscreteTimeDomain(Domain):
    domain = 'discrete time'    
    domain_label = 'Discrete time'
    domain_units = ''
    is_discrete_time_domain = True
    is_transform_domain = False    
    

class DiscreteFourierDomain(Domain):
    domain = 'discrete fourier'
    domain_label = 'Discrete Fourier'    
    domain_units = ''
    is_discrete_fourier_domain = True
    is_transform_domain = True    
    
    
class ZDomain(Domain):
    domain = 'Z'
    domain_label = 'Z'    
    domain_units = ''
    is_Z_domain = True
    is_transform_domain = True    


class SuperpositionDomain(Domain):
    domain = 'superposition'
    domain_label = 'Superposition'    
    domain_units = ''
    is_superposition_domain = True
    is_transform_domain = True    
    
    



#!/usr/bin/env python
"""schtex V0.2
Copyright (c) 2014--2020 Michael P. Hayes, UC ECE, NZ

Usage: schtex infile.sch [outfile.tex|pdf|png|svg|sch|pgf]
"""

from __future__ import print_function
from matplotlib.pyplot import show
from argparse import ArgumentParser
import sys
import os

def schtex_exception(type, value, tb):
   if hasattr(sys, 'ps1') or not sys.stderr.isatty():
      # We are not in interactive mode or we don't have a tty-like
      # device, so call the default hook
      sys.__excepthook__(type, value, tb)
   else:
      import traceback, pdb
      # We are in interactive mode, print the exception...
      traceback.print_exception(type, value, tb)
      print()
      # ...then start the debugger in post-mortem mode.
      pdb.pm()

def main (argv=None):

    if argv is None:
        argv = sys.argv

    version = __doc__.split('\n')[0]

    parser = ArgumentParser(description='Convert Lcapy schematic to image file.')
    parser.add_argument('--version', action='version', version=__doc__.split('\n')[0])
    
    parser.add_argument('--debug', action='store_true',
                        dest='debug', default=None,
                        help="enable debugging")
    
    parser.add_argument('--draw-nodes', type=str,
                        dest='draw_nodes', default=None,
                        help='draw nodes, choice: none, connections, primary, all')
    
    parser.add_argument('--nodraw-nodes', action='store_false',
                        dest='draw_nodes',
                        help="don't draw nodes")
    
    parser.add_argument('--label-nodes', type=str,
                        dest='label_nodes', default=None,
                        help='label nodes, choice: none, alpha, pins, primary, all or a list of comma separated node names in braces, e.g, {1, 2, 3}')
    
    parser.add_argument('--nolabel-nodes', action='store_false',
                        dest='label_nodes',
                        help="don't label nodes")
    
    parser.add_argument('--nolabel-ids', action='store_false',
                        dest='label_ids', default=None,
                        help="don't label element identifiers")
    
    parser.add_argument('--label-ids', action='store_true',
                        dest='label_ids', default=None,
                        help="label element identifiers")
    
    parser.add_argument('--nolabel-values', action='store_false',
                        dest='label_values', default=None,
                        help="don't label values")
    
    parser.add_argument('--label-values', action='store_true',
                        dest='label_values', default=None,
                        help="labels values")
    
    parser.add_argument('--s-model', action='store_true',
                        dest='s_model', default=False,
                        help='generate s-domain model schematic')

    parser.add_argument('--noisy-model', action='store_true',
                        dest='noisy_model', default=False,
                        help='generate noisy model schematic')
    
    parser.add_argument('--ac-model', action='store_true',
                        dest='ac_model', default=False,
                        help='generate AC model schematic')
    
    parser.add_argument('--p-model', action='store_true',
                        dest='p_model', default=False,
                        help='generate pre-initial model schematic')
    
    parser.add_argument('--k-model', action='store_true',
                        dest='k_model', default=False,
                        help='generate schematic with independent sources killed')

    parser.add_argument('--kill-zero', action='store_true',
                        default=False,
                        help='generate schematic with zero independent sources killed')    
    
    parser.add_argument('--scale', type=float,
                        dest='scale', default=None,
                        help='schematic scale factor, this scales the schematic size but not the fonts')
    
    parser.add_argument('--node-spacing', type=float,
                        dest='node_spacing', default=None,
                        help='this specifies the spacing of the nodes of a component')
    
    parser.add_argument('--cpt-size', type=float,
                        dest='cpt_size', default=None,
                        help='this specifies the size of a component; it needs to be smaller than node_spacing')

    parser.add_argument('--dpi', type=float, default=None,
                        help='this specifies the number of dots per inch when creating a png file')    
    
    parser.add_argument('--help-lines', type=float,
                        dest='help_lines', default=None,
                        help="draw help lines")
    
    parser.add_argument('--xgraph', action='store_true',
                        dest='xgraph', default=False,
                        help="generate graph of component x positions")

    parser.add_argument('--circuitgraph', action='store_true',
                        default=False,
                        help="generate circuit graph")    
    
    parser.add_argument('--ygraph', action='store_true',
                        dest='ygraph', default=False,
                        help="generate graph of component y positions")
    
    parser.add_argument('--stage', type=int,
                        dest='stage', default=0,
                        help='select graph analysis stage 0--4')
    
    parser.add_argument('--pdb', action='store_true',
                        default=False,
                        help="enter python debugger on exception")

    parser.add_argument('--allow-anon', action='store_true',
                        default=False,
                        help="allow anonymous component ids")        
    
    parser.add_argument('--renumber', type=str,
                        dest='renumber', default=None,
                        help='renumber nodes, e.g, 10:1, 11:2 or all')

    parser.add_argument('--font', type=str, default=None,
                        help=r'specify default font, e.g., \small, \sffamily\tiny')    

    parser.add_argument('--options', type=str, default=None,
                        help='specify circuitikz options')
    
    parser.add_argument('--preamble', type=str, default=None,
                        help='specify circuitikz commands at start')

    parser.add_argument('--postamble', type=str, default=None,
                        help='specify circuitikz commands at end')

    parser.add_argument('filename', type=str, help='schematic filename')
    parser.add_argument('outfilename', type=str, nargs='?',
                        help='output filename', default=None)    

    parser.add_argument('--show', action='store_true',
                        dest='show', default=None,
                        help="show image")
    
    args = parser.parse_args()    

    infilename = args.filename
    outfilename = args.outfilename

    if args.pdb:
        sys.excepthook = schtex_exception
        
    from lcapy import Circuit, CircuitGraph

    cct = Circuit(infilename, allow_anon=args.allow_anon)
    if args.k_model:
        cct = cct.kill()
    if args.kill_zero:
        cct = cct.kill_zero()        
    if args.s_model:
        cct = cct.s_model()
    if args.ac_model:
        cct = cct.ac_model()
    if args.p_model:
        cct = cct.pre_initial_model()
    if args.noisy_model:
        cct = cct.noisy()        

    if args.renumber == 'all':
       cct = cct.renumber()
    elif args.renumber:
        parts = args.renumber.split(',')
        node_map = {}
        for part in parts:
            part = part.strip()
            fields = part.split(':')
            if len(fields) != 2:
                raise ValueError('Expecting mapping of form a:b got %s' % part)
            node_map[fields[0]] = fields[1]
        cct = cct.renumber(node_map)

    if args.label_nodes not in ('none', 'all', 'alpha', 'pins', 'primary', False, None):
        raise ValueError('Illegal option %s for label_nodes' % args.label_nodes)

    if args.draw_nodes not in ('none', 'all', 'primary', 'connections',
                                  False, None):
        raise ValueError('Illegal option %s for draw_nodes' % args.draw_nodes)

    if outfilename is not None:
       base, ext = os.path.splitext(outfilename)
       if ext == '.sch':
          open(outfilename, 'w').write(str(cct))
          return 0
        
    nosave = args.xgraph or args.ygraph

    if not args.xgraph and not args.ygraph and not args.circuitgraph:
        kwargs = {}
        if args.options != None:
            kwargs['options'] = args.options
        if args.preamble != None:
            kwargs['preamble'] = args.preamble
        if args.postamble != None:
            kwargs['postamble'] = args.postamble            
        if args.font != None:
            kwargs['font'] = args.font          
       
        cct.draw(label_nodes=args.label_nodes,
                 draw_nodes=args.draw_nodes,
                 label_ids=args.label_ids,
                 label_values=args.label_values, 
                 filename=outfilename, scale=args.scale,
                 node_spacing=args.node_spacing, cpt_size=args.cpt_size,
                 help_lines=args.help_lines, debug=args.debug,
                 dpi=args.dpi, nosave=nosave, **kwargs)

    if args.xgraph:
        cct.sch.make_graphs(debug=args.debug)
        cct.sch.xgraph.dot(outfilename, stage=args.stage)

    if args.ygraph:
        cct.sch.make_graphs(debug=args.debug)
        cct.sch.ygraph.dot(outfilename, stage=args.stage)

    if args.circuitgraph:
       cg = CircuitGraph(cct)
       cg.draw(filename=outfilename)

    if args.show:
        show()
        
    return 0


if __name__ == '__main__':
    sys.exit(main())



"""This module provides the ComponentNamer class for automatic
naming of components.

Copyright 2020 Michael Hayes, UCECE

"""


class ComponentNamer(object):

    def __init__(self):

        self.names = {}

    def make(self, cpt_kind, name_kind='anon'):

        if name_kind not in self.names:
            self.names[name_kind] = {}

        foo = self.names[name_kind]
            
        if cpt_kind not in foo:
            foo[cpt_kind] = 0

        foo[cpt_kind] += 1

        return cpt_kind + name_kind + str(foo[cpt_kind])
    
    def __call__(self, cpt_kind, name_kind='anon'):
        """Create component name based on `cpt_kind` (`R`, `L`, etc.)
        and the name_kind (`anon`, `t`, etc)."""

        return self.make(cpt_kind, name_kind)



"""This module performs transformations between domains.

Copyright 2018--2020 Michael Hayes, UCECE

"""

from .sym import sympify, pi
from .symbols import f, s, t, omega, j, jw, jw0
from .expr import expr as expr1
from .expr import Expr


def transform1(expr, arg, **assumptions):

    # Handle things like (3 * s)(5 * s)
    if isinstance(expr, arg.__class__) and not isinstance(expr, Superposition):
        return expr.subs(arg)

    # Handle expr(t), expr(s), expr(f), expr(omega)
    if arg is t:
        return expr.time(**assumptions)
    elif arg is s:
        return expr.laplace(**assumptions)
    elif arg is f:
        return expr.fourier(**assumptions)
    elif arg is omega:
        return expr.angular_fourier(**assumptions)
    elif arg.has(j):
        return expr.phasor(omega=arg / j, **assumptions)    

    # Handle expr(texpr), expr(sexpr), expr(fexpr), expr(omegaexpr).  For example,
    # expr(2 * f).
    result = None 
    if isinstance(arg, TimeDomainExpression):
        result = expr.time(**assumptions)
    elif isinstance(arg, LaplaceDomainExpression):
        result = expr.laplace(**assumptions)
    elif isinstance(arg, FourierDomainExpression):
        result = expr.fourier(**assumptions)
    elif isinstance(arg, AngularFourierDomainExpression):
        result = expr.angular_fourier(**assumptions)        
    elif arg.has(j):
        result = expr.phasor(omega=arg / j, **assumptions)
    elif arg.is_constant:
        if not isinstance(expr, Superposition):
            result = expr.time(**assumptions)
        else:
            result = expr
    else:
        raise ValueError('Can only return t, f, s, omega, or jw domains')

    return result.subs(arg, **assumptions)


def transform(expr, arg, **assumptions):
    """If arg is f, s, t, omega, or jw perform domain transformation,
    otherwise perform substitution.

    Note (1 / s)(omega) will fail since 1 / s is assumed not to be
    causal and so the Fourier transform is unknown.  However,
    impedance(1 / s)(omega) will work since an impedance is assumed to
    be causal.  Alternatively, use (1 / s)(omega, causal=True). """

    arg = expr1(arg)

    new = transform1(expr, arg, **assumptions)
    return new


def call(expr, arg, **assumptions):

    if id(arg) in (id(f), id(s), id(t), id(omega), id(jw), id(jw0)):
        return expr.transform(arg, **assumptions)

    if arg in (f, s, t, omega, jw, jw0):
        return expr.transform(arg, **assumptions)

    try:
        # arg might be an int, float, complex, etc.
        if arg.has(j):
            return expr.transform(arg, **assumptions)
    except:
        pass
        
    return expr.subs(arg)


def select(expr, kind):

    if not isinstance(kind, str):
        return expr.subs(j * kind)                

    # If kind is an expr, then will add 'dc', 'time', etc. as symbols.
    
    if kind == 't':
        return expr.time()
    elif kind in ('dc', 'time'):
        return expr.subs(0)
    elif kind in ('s', 'ivp', 'super', 'laplace'):
        return expr.laplace()
    elif kind == 'f':
        return expr.fourier()
    elif kind == 'omega':
        return expr.angular_fourier()
    elif isinstance(kind, str) and kind.startswith('n'):
        return expr.angular_fourier()
    else:
        raise RuntimeError('unknown kind')


from .fexpr import FourierDomainExpression    
from .sexpr import LaplaceDomainExpression
from .texpr import TimeDomainExpression
from .omegaexpr import AngularFourierDomainExpression
from .super import Superposition
from .current import current
from .voltage import voltage
from .admittance import admittance
from .impedance import impedance
from .transfer import transfer



from lcapy import *

a = Circuit('VRC2.sch')

H = a.P1.transfer('P2')

H(jw).dB.plot((0.1, 10), log_frequency=True)

from matplotlib.pyplot import savefig
savefig(__file__.replace('.py', '.png'), bbox_inches='tight')








"""This module provides admittance support.

Copyright 2019-2020 Michael Hayes, UCECE

"""
from __future__ import division
from .expr import expr
from .sexpr import LaplaceDomainExpression
from .units import u as uu

    
def admittance(arg, **assumptions):
    """Generic admittance factory function.

    Y(omega) = G(omega) + j * B(omega)

    where G is the conductance and B is the susceptance.

    Admittance is the reciprocal of impedance,

    Z(omega) = 1 / Y(omega)

    """    

    expr1 = expr(arg, **assumptions)

    try:
        expr1 = expr1.as_admittance()
    except:
        raise ValueError('Cannot represent %s(%s) as admittance' % (expr1.__class__.__name__, expr1))        

    # Could use siemens but this causes comparison problems if
    # unit simplification not performed.
    return expr1.apply_unit(1 / uu.ohms)




from lcapy import Circuit, expr
import unittest
import sympy as sym


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy state space analysis

    """

    def assertEqual2(self, ans1, ans2, comment):

        ans1 = ans1.canonical()
        ans2 = ans2.canonical()

        print(ans1, ans2)
        try:
            self.assertEqual(ans1, ans2, comment)
        except AssertionError as e:
            ans1.pprint()
            ans2.pprint()
            raise AssertionError(e)

    def test_VRC1(self):
        """Lcapy: check VRC circuit

        """
        a = Circuit("""
        V1 1 0
        R1 1 2
        C1 2 0""")

        ss = a.ss
        
        self.assertEqual2(expr(ss.x[0]), expr('v_C1(t)'), "Incorrect state variable")
        self.assertEqual2(expr(ss.y[0]), expr('v_1(t)'), "Incorrect output variable1")
        self.assertEqual2(expr(ss.y[1]), expr('v_2(t)'), "Incorrect output variable2")        
        self.assertEqual2(expr(ss.A[0]), expr('-1/(R1 * C1)'), "Incorrect A matrix")
        self.assertEqual2(expr(ss.B[0]), expr('1/(R1 * C1)'), "Incorrect B matrix")
        self.assertEqual2(expr(ss.C[0]), expr(0), "Incorrect C[0] matrix element")
        self.assertEqual2(expr(ss.C[1]), expr(1), "Incorrect C[1] matrix element")
        self.assertEqual2(expr(ss.D[0]), expr(1), "Incorrect D[0] matrix element")
        self.assertEqual2(expr(ss.D[1]), expr(0), "Incorrect D[1] matrix element")
        self.assertEqual2(expr(ss.eigenvalues[0]), expr('-1/(R1 * C1)'), "Incorrect eigenvalues")
        
        
    def test_VRL1(self):
        """Lcapy: check VRL circuit

        """
        a = Circuit("""
        V1 1 0
        R1 1 2
        L1 2 0""")

        ss = a.ss
        
        self.assertEqual2(expr(ss.x[0]), expr('i_L1(t)'), "Incorrect state variable")
        self.assertEqual2(expr(ss.y[0]), expr('v_1(t)'), "Incorrect output variable1")
        self.assertEqual2(expr(ss.y[1]), expr('v_2(t)'), "Incorrect output variable2")        
        self.assertEqual2(expr(ss.A[0]), expr('-R1 / L1'), "Incorrect A matrix")
        self.assertEqual2(expr(ss.B[0]), expr('1 / L1'), "Incorrect B matrix")
        self.assertEqual2(expr(ss.C[0]), expr(0), "Incorrect C[0] matrix element")
        self.assertEqual2(expr(ss.C[1]), expr('-R1'), "Incorrect C[1] matrix element")
        self.assertEqual2(expr(ss.D[0]), expr(1), "Incorrect D[0] matrix element")
        self.assertEqual2(expr(ss.D[1]), expr(1), "Incorrect D[1] matrix element")
        self.assertEqual2(expr(ss.eigenvalues[0]), expr('-R1 / L1'), "Incorrect eigenvalues")
        
